import os
import shutil
import sys
import logging

import depthmap
import exporter

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)


if __name__ == "__main__":
    if len(sys.argv) != 3:
        print('You did not enter depthmap_dir folder and calibration file path')
        print('E.g.: python convertdepth2pcd.py depthmap_dir calibration file')
        sys.exit(1)

    depthmap_dir = sys.argv[1]
    calibration_file = sys.argv[2]

    depth_filenames = []
    for (dirpath, dirnames, filenames) in os.walk(depthmap_dir + '/depth'):
        depth_filenames.extend(filenames)
    depth_filenames.sort()
    try:
        shutil.rmtree('export')
    except BaseException:
        print('no previous data to delete')
    os.mkdir('export')
    for filename in depth_filenames:
        dmap = depthmap.Depthmap.create_from_zip(depthmap_dir, filename, 0, calibration_file)
        output_filename = f'export/output{filename}.pcd'
        exporter.export_pcd(output_filename, dmap)

    logger.info('Data exported into folder export')



import pickle
from typing import List, Tuple, Union
import logging

import numpy as np
import tensorflow as tf

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)


def create_multiartifact_sample(artifacts: List[str],
                                normalization_value: float,
                                image_target_height: int,
                                image_target_width: int,
                                targets_indices: tf.Tensor,
                                n_artifacts: int,
                                ) -> Tuple[tf.Tensor, tf.Tensor]:
    """Open pickle files and load data.

    Args:
        artifacts: List of file paths to pickle files

    Returns:
        depthmaps of shape (IMAGE_TARGET_HEIGHT, IMAGE_TARGET_WIDTH, n_artifacts)
        targets of shape (1, )
    """
    targets_list = []
    depthmaps = np.zeros((image_target_height, image_target_width, n_artifacts))

    for i, artifact_path in enumerate(artifacts):
        depthmap, targets = _py_load_pickle(artifact_path, normalization_value,
                                            image_target_height, image_target_width, targets_indices)
        depthmap.set_shape((image_target_height, image_target_width, 1))
        depthmaps[:, :, i] = tf.squeeze(depthmap, axis=2)
        targets_list.append(targets)
    targets = targets_list[0]
    if not np.all(targets_list == targets):
        logger.info('Warning: Not all targets are the same!! \n target_list: %s \n artifacts: %s: ',
                    targets_list, artifacts)
    return depthmaps, targets


def _py_load_pickle(path: Union[tf.Tensor, str],
                    normalization_value: float,
                    image_target_height: int,
                    image_target_width: int,
                    targets_indices: tf.Tensor,
                    ) -> Tuple[Union[tf.Tensor, np.array], Union[tf.Tensor, np.array]]:
    path_ = path if isinstance(path, str) else path.numpy()
    try:
        depthmap, targets = pickle.load(open(path_, "rb"))
    except OSError as e:
        print(f"path: {path}, type(path) {str(type(path))}")
        print(e)
        raise e
    depthmap = _preprocess_depthmap(depthmap)
    depthmap = depthmap / normalization_value
    depthmap = tf.image.resize(depthmap, (image_target_height, image_target_width))
    targets = _preprocess_targets(targets, targets_indices)
    return depthmap, targets


def _preprocess_depthmap(depthmap: Union[tf.Tensor, np.array]) -> Union[tf.Tensor, np.array]:
    return depthmap.astype("float32")


def _preprocess_targets(targets: Union[tf.Tensor, np.array],
                        targets_indices_: tf.Tensor,
                        ) -> Union[tf.Tensor, np.array]:
    targets_indices = targets_indices_.numpy().tolist()
    if targets_indices is not None:
        targets = targets[targets_indices]
    return targets.astype("float32")



import numpy as np
from PIL import Image
from torchvision import transforms

from .coco import NAMES

# each class has different color bbox
COLORS = [tuple(c) for c in np.random.randint(0, 255, size=(len(NAMES), 3))]


def predict(image, model, threshold=0.8):
    """Return dict of boxes, scores, labels (class index), names, masks"""
    model.eval()
    transform = transforms.ToTensor()
    image = transform(image)
    image = image.unsqueeze(0)
    out = model(image)[0]

    index = (out["scores"] >= threshold) & (out["labels"] == NAMES.index("person"))
    for k, v in out.items():
        out[k] = v.detach().cpu().numpy()[index]
    out["names"] = [NAMES[i] for i in out["labels"]]
    return out


def apply_mask(image, mask, color, alpha=0.5):
    """Apply the given mask to the image """
    image = np.array(image)
    for c in range(3):
        image[:, :, c] = np.where(
            mask == 1,
            image[:, :, c] * (1 - alpha) + alpha * color[c] * 255,
            image[:, :, c],
        )
    return Image.fromarray(image)


def show(image, out, alpha=0.9999):
    """Display image with boxes, scores, names, masks """
    image = image.copy()
    masks = out.get("masks", [])

    for mask in masks:
        # mask fill. random color to split items
        mask = mask[0].round().astype(np.uint8)
        color = tuple([255, 255, 255])
        image = apply_mask(image, mask, color, alpha)

    return image



#
# Child Growth Monitor - Free Software for Zero Hunger
# Copyright (c) 2019 Dr. Christian Pfitzner <christian.pfitzner@th-nuernberg.de> for Welthungerhilfe
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os
from typing import Iterable

import numpy as np
from cv2 import cv2
from PIL import Image
from pyntcloud import PyntCloud

from cgmml.common.rgbd_toolkit.cgm_fusion.calibration import get_extrinsic_matrix, get_intrinsic_matrix, get_k
from cgmml.common.rgbd_toolkit.cgm_fusion.utility import fuse_point_cloud

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)


def project_points(pcd_points, calibration_file):

    #get the data for calibration
    intrinsic = get_intrinsic_matrix(calibration_file)
    ext_d = get_extrinsic_matrix(calibration_file, idx=4)

    r_vec = ext_d[:3, :3]
    t_vec = -ext_d[:3, 3]

    k1, k2, k3 = get_k(calibration_file)

    im_coords, jac = cv2.projectPoints(pcd_points, r_vec,
                                       t_vec, intrinsic[:3, :3],
                                       np.array([k1, k2, 0, 0]))

    return im_coords, jac


def get_depth_image_from_point_cloud(calibration_file, pcd_file, output_file):
    if not os.path.exists(pcd_file):  # check all files exist
        logger.info('Point cloud does not exist')
        return

    if not os.path.exists(calibration_file):  # check if the califile exists
        logger.info('Calibration does not exist')
        return

    try:
        cloud = PyntCloud.from_file(pcd_file)  # load the data from the files
    except ValueError:
        logger.info(" Error reading point cloud ")
        raise

    # points       = cloud.points.values[:, :3]
    z = cloud.points.values[:, 3]

    logger.info(cloud.points.values.shape)

    #height = 172  # todo: get this from calibration file

    z = (z - min(z)) / (max(z) - min(z))  # normalize the data to 0 to 1

    # logger.info(z)

    # logger.info(z.size)

    # iterat of the points and calculat the x y coordinates in the image
    # get the data for calibration
    # im_coords = apply_projection(points)

    # manipulate the pixels color value depending on the z coordinate
    # for i, t in enumerate(im_coords):
    #     x, y = t.squeeze()
    #     x = int(np.round(x))
    #     y = int(np.round(y))
    #     if x >= 0 and x < width and y >= 0 and y < height:
    #         viz_image[x,y] = 255*z[i]

    # # resize and  return the image after pricessing
    # imgScale  = 0.25
    # newX,newY = viz_image.shape[1]*imgScale, viz_image.shape[0]*imgScale
    # cv2.imwrite('/tmp/depth_visualization.png', viz_image)

    #111 utilit get vizc_channel

    np.resize(z * 255, [224, 172, 3])

    # depth_img_resize = cv2.resize(
    #     z * 255, (180, 180))  # todo: make width and height variable

    #cv2.imwrite("/tmp/depth_224x172.png", depth_img)
    #cv2.imwrite("/tmp/depth_240x180.png", depth_img_resize)

    # not sure if we need this

    # # get the data for calibration
    # intrinsic  = get_intrinsic_matrix()
    # ext_d      = get_extrinsic_matrix(4)

    # r_vec      = ext_d[:3, :3]
    # t_vec      = -ext_d[:3, 3]

    # k1, k2, k3 = get_k()
    # im_coords, _ = cv2.projectPoints(points, r_vec, t_vec, intrinsic[:3, :3], np.array([k1, k2, 0, 0]))


def does_path_belong_to_codes(path: str, codes: Iterable) -> bool:
    for code in codes:
        if f"_{code}_" in path:
            return True
    return False


def fuse_rgbd(calibration_file,
              pcd_file,
              image,
              seg_path=0):  # TODO:uncomment zero for segmentation fusion

    try:
        cloud = PyntCloud.from_file(pcd_file)  # load the data from the files
    except ValueError:
        logger.error(" Error reading point cloud ")
        raise

    points = cloud.points.values[:, :3]
    #getting projected points
    im_coords, jac = project_points(points, calibration_file)

    #setting the RGB image dimensions
    scale = 0.1
    width = int(1920 * scale)
    height = int(1080 * scale)
    #reading image and resizing
    pil_im = image.resize((width, height), Image.ANTIALIAS)
    im_array = np.asarray(pil_im)

    #TODO:comment out for segmentation
    #pil_im2 = Image.open(seg_path)
    #pil_im2 = pil_im2.resize((height, width), Image.ANTIALIAS)
    #im_array2 = np.asarray(pil_im2)
    #segm = im_array2[y][x][1] / 255.0

    #initialize an empty black image
    viz_image = np.zeros((width, height, 4))

    #addING depth into RGB array
    pcd_name = pcd_file.split("/")[-1]

    for i in range(len(points)):
        x = int(im_coords[i][0][0] * scale)
        y = int(im_coords[i][0][1] * scale)
        if x >= 0 and y >= 0 and x < width and y < height:
            depth = points[i][2]

            if does_path_belong_to_codes(pcd_name, codes=("100", "101", "102")):
                newx = y
                newy = width - x - 1

            elif does_path_belong_to_codes(pcd_name, codes=("200", "201", "202")):
                newx = height - y - 1
                newy = x
            else:
                raise NameError(f"{pcd_name} does not have a correct code")

            viz_image[newy][height - newx - 1][3] = depth

    for x in range(width):
        for y in range(height):

            if does_path_belong_to_codes(pcd_name, codes=("100", "101", "102")):
                newx = y
                newy = width - x - 1

            elif does_path_belong_to_codes(pcd_name, codes=("200", "201", "202")):
                newx = height - y - 1
                newy = x
            else:
                raise NameError(f"{pcd_name} does not have a correct code")

            viz_image[newy][newx][0] = im_array[y][x][0] / 255.0
            viz_image[newy][newx][1] = im_array[y][x][1] / 255.0
            viz_image[newy][newx][2] = im_array[y][x][2] / 255.0
    return viz_image


def apply_fusion(calibration_file, pcd_file, jpg_file, seg_path):
    """Check the path if everything is correct"""
    if not os.path.exists(pcd_file):  # check all files exist
        logger.error('Point cloud does not exist')
        return

    if not os.path.exists(jpg_file):  # check if the jpg file exists
        logger.error('Image does not exist')
        return

    if not os.path.exists(seg_path):  # check if segmentation exists
        logger.error('Segmentation not found')
        return

    if not os.path.exists(calibration_file):  # check if the califile exists
        logger.error('Calibration does not exist')
        return

    try:
        cloud = PyntCloud.from_file(pcd_file)  # load the data from the files
    except ValueError:
        logger.error(" Error reading point cloud ")
        raise

    jpg = cv2.imread(jpg_file, -1)
    jpg = cv2.flip(jpg, 0)

    seg = cv2.imread(seg_path, -1)
    seg = cv2.flip(seg, 0)

    hh, ww, _ = jpg.shape

    points = cloud.points.values[:, :3]

    confidence = cloud.points.values[:, 3]

    # get the data for calibration

    im_coords, _ = project_points(points, calibration_file)

    color_vals = np.zeros_like(points)

    segment_vals = np.zeros_like(points)

    for i, t in enumerate(im_coords):
        x, y = t.squeeze()
        x = int(np.round(x))
        y = int(np.round(y))
        if x >= 0 and x < ww and y >= 0 and y < hh:
            color_vals[i, :] = jpg[y, x]
            segment_vals[i, :] = seg[y, x]

    # #convert from pyntcloud to open3d
    # cloud_open3d = o3d.io.read_point_cloud(pcd_file)

    # #calculate the normals from the existing cloud
    # cloud_open3d.estimate_normals(search_param=o3d.geometry.KDTreeSearchParamHybrid(radius=0.1, max_nn=30))

    fused_point_cloud = fuse_point_cloud(
        points, color_vals, confidence,
        segment_vals)  # , np.asarray(cloud_open3d.normals))

    return fused_point_cloud



import os
import re
import logging
from functools import partial
from itertools import groupby, islice
from typing import Iterator, List

from bunch import Bunch
import glob2 as glob

from cgmml.common.model_utils.model_utils_constants import SAMPLING_STRATEGY_SYSTEMATIC, SAMPLING_STRATEGY_WINDOW

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)

REGEX_PICKLE = re.compile(
    r"pc_(?P<qrcode>[a-zA-Z0-9]+-[a-zA-Z0-9]+)_(?P<unixepoch>\d+)_(?P<code>\d{3})_(?P<idx>\d{3}).p$"
)


def create_multiartifact_paths_for_qrcodes(qrcode_paths: List[str], data_config: Bunch) -> List[List[str]]:
    samples = []
    for qrcode_path in sorted(qrcode_paths):
        for code in data_config.CODES:
            p = os.path.join(qrcode_path, code)
            new_samples = _create_multiartifact_paths(p, data_config.N_ARTIFACTS, data_config)
            samples.extend(new_samples)
    return samples


def _create_multiartifact_paths(qrcode_path: str, n_artifacts: int, data_config: Bunch) -> List[List[str]]:
    """Look at files for 1 qrcode and divide into samples.

    Args:
        qrcode_path: File path of 1 qrcode, e.g. "dataset/scans/1583462470-16tvfmb1d0/100"
        n_artifacts: Desired number of artifacts in one sample

    Returns:
        List of samples, where each sample consists of muliple file paths
    """
    path_with_wildcard = os.path.join(qrcode_path, "*.p")
    list_of_pickle_file_paths = sorted(glob.glob(path_with_wildcard))

    # Split if there are multiple scans on different days
    scans = [list(v) for _unixepoch, v in groupby(list_of_pickle_file_paths, _get_epoch) if _unixepoch]

    # Filter to keep scans with enough artifacts
    scans = list(filter(lambda x: len(x) >= n_artifacts, scans))

    # Sample artifacts
    if data_config.SAMPLING_STRATEGY == SAMPLING_STRATEGY_SYSTEMATIC:
        samples = list(map(partial(sample_systematic_from_artifacts, n_artifacts=n_artifacts), scans))

    if data_config.SAMPLING_STRATEGY == SAMPLING_STRATEGY_WINDOW:
        samples = []
        for scan in scans:
            some_samples = list(sample_windows_from_artifacts(scan, n_artifacts=n_artifacts))
            assert len(scan) - n_artifacts + 1 == len(some_samples)
            samples.extend(some_samples)

    return samples


def sample_windows_from_artifacts(artifacts: list, n_artifacts: int) -> Iterator[list]:
    """Sample multiple windows (of length n_artifacts) from list of artifacts

    Args:
        artifacts: e.g. ['001.p', '002.p', '003.p', '004.p', '005.p', '006.p']
        n_artifacts: Desired number of artifacts in one sample

    Returns:
        samples: e.g. [
            ['001.p', '002.p', '003.p', '004.p', '005.p'],
            ['002.p', '003.p', '004.p', '005.p', '006.p'],
        ]
    """
    it = iter(artifacts)
    result = list(islice(it, n_artifacts))
    if len(result) == n_artifacts:
        yield result
    for elem in it:
        result = result[1:] + [elem]
        yield result


def sample_systematic_from_artifacts(artifacts: list, n_artifacts: int) -> list:
    n_artifacts_total = len(artifacts)
    n_skip = n_artifacts_total // n_artifacts  # 20 / 5 = 4
    indexes_to_select = list(range(n_skip // 2, n_artifacts_total, n_skip))[:n_artifacts]
    selected_artifacts = [artifacts[i] for i in indexes_to_select]
    assert len(selected_artifacts) == n_artifacts, str(artifacts)
    return selected_artifacts


def _get_epoch(fname: str) -> str:
    match_result = REGEX_PICKLE.search(fname)
    if match_result:
        return match_result.group("unixepoch")
    else:
        logger.info("%s doesn't match REGEX_PICKLE", fname)



import os
import logging
from pathlib import Path
import subprocess

import tensorflow as tf
import tensorflow_addons as tfa
from azureml.core.run import Run
from azureml.core.workspace import Workspace
from tensorflow.keras import callbacks

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)


def get_optimizer(use_one_cycle: bool, lr: float, n_steps: int):
    if use_one_cycle:
        lr_schedule = tfa.optimizers.TriangularCyclicalLearningRate(
            initial_learning_rate=lr / 100,
            maximal_learning_rate=lr,
            step_size=n_steps,
        )
        # Note: When using 1cycle, this uses the Adam (not Nadam) optimizer
        return tf.keras.optimizers.Adam(learning_rate=lr_schedule)
    return tf.keras.optimizers.Nadam(learning_rate=lr)


def download_dataset(workspace: Workspace, dataset_name: str, dataset_path: str):
    logger.info("Accessing dataset...")
    if os.path.exists(dataset_path):
        return
    dataset = workspace.datasets[dataset_name]
    logger.info("Downloading dataset %s", dataset_name)
    dataset.download(target_path=dataset_path, overwrite=False)
    logger.info("Finished downloading %s", dataset_name)


def get_dataset_path(data_dir: Path, dataset_name: str) -> str:
    return str(data_dir / dataset_name)


class AzureLogCallback(callbacks.Callback):
    """Pushes metrics and losses into the run on AzureML"""
    def __init__(self, run: Run):
        super().__init__()
        self.run = run

    def on_epoch_end(self, epoch, logs=None):
        if logs is not None:
            for key, value in logs.items():
                self.run.log(key, value)


def create_tensorboard_callback() -> callbacks.TensorBoard:
    return callbacks.TensorBoard(
        log_dir="logs",
        histogram_freq=0,
        write_graph=True,
        write_grads=False,
        write_images=True,
        embeddings_freq=0,
        embeddings_layer_names=None,
        embeddings_metadata=None,
        embeddings_data=None,
        update_freq="epoch"
    )


WANDB_API_KEY_MH = "237ca046c5dcd915945761dc477207549ef2c42c"


def setup_wandb():
    wandb_login = subprocess.run(["wandb", "login", WANDB_API_KEY_MH])
    assert wandb_login.returncode == 0



"""
This file contains a data-generator that works on the data as it is provided by the ETL.
"""

from __future__ import absolute_import
import os
import logging
import numpy as np
import glob2 as glob
import random
import uuid
import progressbar
import multiprocessing as mp
from pyntcloud import PyntCloud
import pickle
from . import utils

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)


class ETLDataGenerator(object):
    """
    This class generates data for training.
    """

    def __init__(
        self,
        dataset_path,
        input_type,
        #output_targets,
        sequence_length=0,
        image_target_shape=(160, 90),
        voxelgrid_target_shape=(32, 32, 32),
        voxel_size_meters=0.01,
        voxelgrid_random_rotation=False,
        pointcloud_target_size=32000,
        pointcloud_random_rotation=False
    ):
        """
        Initializes a DataGenerator.

        Args:
            dataset_path (string): Where the raw data is.
            input_type (string): Specifies how the input-data for the Neural Network looks like.
                                 Either 'image', 'pointcloud', 'voxgrid'.
            output_targets (list of strings): A list of targets for the Neural Network.
                                 For example *['height', 'weight']*.
            sequence_length (int): Specifies the lenght of the sequences. 0 would yield no sequence at all.
            image_target_shape (2D tuple of ints): Target shape of the images.
            voxelgrid_target_shape (3D tuple of ints): Target shape of the voxelgrids.
            voxel_size_meters (float): Size of the voxels. That is, edge length.
            voxelgrid_random_rotation (bool): If True voxelgrids will be rotated randomly.
            pointcloud_target_size (int): Target size of the pointclouds.
            pointcloud_random_rotation (bool): If True pointclouds will be rotated randomly.

        """

        # Preconditions.
        assert os.path.exists(dataset_path), "dataset_path must exist: " + str(dataset_path)
        assert isinstance(input_type, str), "input_type must be string: " + str(input_type)
        #assert isinstance(output_targets, list), "output_targets must be list: " + str(output_targets)
        if input_type == "image":
            assert len(image_target_shape) == 2, "image_target_shape must be 2-dimensional: " + \
                str(image_target_shape)
        if input_type == "voxelgrid":
            assert len(voxelgrid_target_shape) == 3, "voxelgrid_target_shape must be 3-dimensional: " + \
                str(voxelgrid_target_shape)

        # Assign the instance-variables.
        self.dataset_path = dataset_path
        self.input_type = input_type
        #self.output_targets = output_targets
        self.sequence_length = sequence_length
        self.image_target_shape = image_target_shape
        self.voxelgrid_target_shape = voxelgrid_target_shape
        self.voxel_size_meters = voxel_size_meters
        self.voxelgrid_random_rotation = voxelgrid_random_rotation
        self.pointcloud_target_size = pointcloud_target_size
        self.pointcloud_random_rotation = pointcloud_random_rotation

        # Find all QR-codes.
        self._find_qrcodes()
        assert self.qrcodes is not None, "No QR-codes found!"

        # Prepare the data.
        self._prepare_qrcodes_dictionary()

        # Create some caches.
        self.image_cache = {}
        self.voxelgrid_cache = {}
        self.pointcloud_cache = {}

        # Check if paths are fine.
        if self.input_type == "image":
            assert self.all_jpg_paths != []
        elif self.input_type == "voxelgrid" or self.input_type == "pointcloud":
            assert self.all_pcd_paths != []
        else:
            raise Exception("Unexpected: " + self.input_type)

    def _find_qrcodes(self):
        """
        Finds all QR-codes.

        Each individual is represented via a unique QR-code. This method extracts the set of QR-codes.
        """

        # Retrieve the QR-codes from the folders.
        paths = glob.glob(os.path.join(self.dataset_path, "*"))
        self.qrcodes = sorted([path.split("/")[-1] for path in paths])

    def _prepare_qrcodes_dictionary(self):

        self.all_pcd_paths = []
        self.all_jpg_paths = []

        self.qrcodes_dictionary = {}
        for qrcode in self.qrcodes:
            self.qrcodes_dictionary[qrcode] = []
            #logger.info("QR-code:", qrcode)
            measurement_paths = glob.glob(os.path.join(self.dataset_path, qrcode, "*"))
            for measurement_path in measurement_paths:
                # Getting PCDs.
                pcd_paths = glob.glob(os.path.join(measurement_path, "pcd", "*.pcd"))

                # Getting JPGs.
                jpg_paths = glob.glob(os.path.join(measurement_path, "jpg", "*.jpg"))

                # Loading the targets.
                target_path = os.path.join(measurement_path, "target.txt")
                target_file = open(target_path, "r")
                targets = np.array([float(value) for value in target_file.read().split(",")])
                target_file.close()

                # Done.
                self.qrcodes_dictionary[qrcode].append((pcd_paths, jpg_paths, targets))
                self.all_pcd_paths.extend(pcd_paths)
                self.all_jpg_paths.extend(jpg_paths)

    def analyze_files(self):

        for qrcode in self.qrcodes:
            logger.info("QR-code: %s", qrcode)
            for pcd_paths, jpg_paths, targets in self.qrcodes_dictionary[qrcode]:
                logger.info(" %d PCD-files for targets %s", len(pcd_paths), targets)
                logger.info(" %d JPG-files for targets %s", len(jpg_paths), targets)

        logger.info("Total PCD-files: %d", len(self.all_pcd_paths))
        logger.info("Total JPG-files: %d", len(self.all_jpg_paths))

        #measurement_paths = glob.glob(os.path.join(self.dataset_path, qrcode, "*"))
        #for measurement_path in measurement_paths:
        #    logger.info("  %s", measurement_path.split("/")[-1])
        #    pcd_paths = glob.glob(os.path.join(measurement_path, "*.pcd"))
        #    logger.info("    %d PCDs", len(pcd_paths))
        #logger.info("Number of JPGs: %d", len(self.jpg_paths))
        #logger.info("Number of PCDs: %d", len(self.pcd_paths))
        #logger.info("Number of JSONs (personal): %d", len(self.json_paths_personal))
        #logger.info("Number of JSONs (measures): %d", len(self.json_paths_measures))

    def generate(self, size, qrcodes_to_use=None, verbose=False, yield_file_paths=False, multiprocessing_jobs=1):

        if qrcodes_to_use is None:
            qrcodes_to_use = self.qrcodes

        # Main loop.
        while True:

            # Use only a single process.
            if multiprocessing_jobs == 1:
                yield generate_data(self, size, qrcodes_to_use, verbose, yield_file_paths, None)

            # Use multiple processes.
            elif multiprocessing_jobs > 1:

                # Create chunks of almost equal size.
                subset_sizes = [0] * multiprocessing_jobs
                subset_sizes[0:multiprocessing_jobs - 1] = [size
                                                            // multiprocessing_jobs] * (multiprocessing_jobs - 1)
                subset_sizes[multiprocessing_jobs - 1] = size - \
                    sum(subset_sizes[0:multiprocessing_jobs - 1])
                subset_sizes = [s for s in subset_sizes if s > 0]
                assert sum(subset_sizes) == size

                # Create an output_queue.
                output_queue = mp.Queue()

                # Create the processes.
                processes = []
                for subset_size in subset_sizes:
                    process_target = generate_data
                    process_args = (self, subset_size, qrcodes_to_use,
                                    verbose, yield_file_paths, output_queue)
                    process = mp.Process(target=process_target, args=process_args)
                    processes.append(process)

                # Start the processes.
                for p in processes:
                    p.start()

                # Exit the completed processes.
                for p in processes:
                    p.join()

                # Get process results from the output queue
                output_paths = [output_queue.get() for p in processes]

                # Merge data.
                x_inputs_arrays = []
                y_outputs_arrays = []
                file_paths_arrays = []
                for output_path in output_paths:
                    # Read data from file and delete it.
                    result_values = pickle.load(open(output_path, "rb"))
                    assert len(result_values[0]) > 0
                    assert len(result_values[1]) > 0
                    os.remove(output_path)

                    # Gather the data into arrays.
                    if x_inputs_arrays != []:
                        assert result_values[0].shape[1:] == x_inputs_arrays[-1].shape[1:], str(
                            result_values[0].shape) + " vs " + str(x_inputs_arrays[-1].shape)
                    if y_outputs_arrays != []:
                        assert result_values[1].shape[1:] == y_outputs_arrays[-1].shape[1:], str(
                            result_values[1].shape) + " vs " + str(y_outputs_arrays[-1].shape)
                    x_inputs_arrays.append(result_values[0])
                    y_outputs_arrays.append(result_values[1])
                    if yield_file_paths is True:
                        file_paths_arrays.append(result_values[2])
                    else:
                        file_paths_arrays.append([])

                x_inputs = np.concatenate(x_inputs_arrays)
                y_outputs = np.concatenate(y_outputs_arrays)
                file_paths = np.concatenate(file_paths_arrays)
                assert len(x_inputs) == size
                assert len(y_outputs) == size

                # Done.
                if yield_file_paths is False:
                    yield x_inputs, y_outputs
                else:
                    yield x_inputs, y_outputs, file_paths

            else:
                raise Exception("Unexpected value for 'multiprocessing_jobs' "
                                + str(multiprocessing_jobs))

    def _load_pointcloud(self, pcd_path, preprocess=True, augmentation=True):

        pointcloud = self.pointcloud_cache.get(pcd_path, [])
        if pointcloud is None:
            pointcloud = PyntCloud.from_file(pcd_path).points.values

            if self.pointcloud_target_size is not None and preprocess is True:
                pointcloud = np.array(pointcloud)[:, 0:3]  # Drop confidence.
                pointcloud = pointcloud[:self.pointcloud_target_size]
                if len(pointcloud) < self.pointcloud_target_size:
                    zeros = np.zeros((self.pointcloud_target_size - len(pointcloud), 4))
                    pointcloud = np.concatenate([pointcloud, zeros])

            if self.pointcloud_random_rotation is True and augmentation is True:
                numpy_points = pointcloud[:, 0:3]
                numpy_points = self._rotate_point_cloud(numpy_points)
                pointcloud[:, 0:3] = numpy_points

            #self.pointcloud_cache[pcd_path] = pointcloud
        return pointcloud

    def _load_voxelgrid(self, pcd_path, preprocess=True, augmentation=True):
        voxelgrid = self.voxelgrid_cache.get(pcd_path, [])
        if voxelgrid == []:

            # Load the pointcloud.
            point_cloud = PyntCloud.from_file(pcd_path)
            if self.voxelgrid_random_rotation is True and augmentation is True:
                points = point_cloud.points
                numpy_points = points.values[:, 0:3]
                numpy_points = self._rotate_point_cloud(numpy_points)
                points.iloc[:, 0:3] = numpy_points
                point_cloud.points = points

            # Create voxelgrid from pointcloud.
            voxelgrid_id = point_cloud.add_structure(
                "voxelgrid",
                size_x=self.voxel_size_meters,
                size_y=self.voxel_size_meters,
                size_z=self.voxel_size_meters)
            voxelgrid = point_cloud.structures[voxelgrid_id].get_feature_vector(mode="density")

            # Do the preprocessing.
            if preprocess is True:
                voxelgrid = utils.ensure_voxelgrid_shape(voxelgrid, self.voxelgrid_target_shape)
                assert voxelgrid.shape == self.voxelgrid_target_shape

            #self.voxelgrid_cache[pcd_path] = voxelgrid # TODO cache is turned off because of you know why...

        return voxelgrid

    def _rotate_point_cloud(self, point_cloud):

        rotation_angle = np.random.uniform() * 2 * np.pi
        cosval = np.cos(rotation_angle)
        sinval = np.sin(rotation_angle)
        rotation_matrix = np.array([[cosval, sinval, 0],
                                    [-sinval, cosval, 0],
                                    [0, 0, 1]])

        rotated_data = np.zeros(point_cloud.shape, dtype=np.float32)
        for k in range(point_cloud.shape[0]):

            shape_pc = point_cloud[k, ...]
            rotated_data[k, ...] = np.dot(shape_pc.reshape((-1, 3)), rotation_matrix)

        return rotated_data


def create_datagenerator_from_parameters(dataset_path, dataset_parameters):
    logger.info("Creating data-generator...")
    datagenerator = ETLDataGenerator(
        dataset_path=dataset_path,
        input_type=dataset_parameters["input_type"],
        #output_targets=dataset_parameters["output_targets"],
        sequence_length=dataset_parameters.get("sequence_length", 0),
        voxelgrid_target_shape=dataset_parameters.get("voxelgrid_target_shape", None),
        voxel_size_meters=dataset_parameters.get("voxel_size_meters", None),
        voxelgrid_random_rotation=dataset_parameters.get("voxelgrid_random_rotation", None),
        pointcloud_target_size=dataset_parameters.get("pointcloud_target_size", None),
        pointcloud_random_rotation=dataset_parameters.get("pointcloud_random_rotation", None)
    )
    #datagenerator.print_statistics()
    return datagenerator


def get_dataset_path(root_path="../data/etl"):
    if os.path.exists("etldatasetpath.txt"):
        with open("etldatasetpath.txt", "r") as file:
            dataset_path = file.read().replace("\n", "")
    else:
        # Finding the latest.
        dataset_paths = glob.glob(os.path.join(root_path, "*"))
        dataset_paths = [
            dataset_path for dataset_path in dataset_paths if os.path.isdir(dataset_path)]
        dataset_path = list(reversed(sorted(dataset_paths)))[0]

    return dataset_path


def generate_data(class_self, size, qrcodes_to_use, verbose, yield_file_paths, output_queue):
    if verbose is True:
        logger.info("Generating QR-codes to be used: %s", qrcodes_to_use)

    assert size != 0

    x_inputs = []
    y_outputs = []
    file_paths = []

    if verbose is True:
        bar = progressbar.ProgressBar(max_value=size)
    while len(x_inputs) < size:

        # Get a random QR-code.
        qrcode = random.choice(qrcodes_to_use)

        # Get targets and paths randomly.
        if qrcode not in class_self.qrcodes_dictionary.keys():
            continue
        pcd_paths, jpg_paths, targets = random.choice(class_self.qrcodes_dictionary[qrcode])

        # No pointclouds.
        if len(pcd_paths) == 0:
            continue

        # Get a sample.
        x_input = None
        y_output = None
        file_path = None

        # Get the input. Not dealing with sequences.
        if class_self.sequence_length == 0:
            x_input, file_path = get_input(class_self, jpg_paths, pcd_paths)

        # Get the input. Dealing with sequences here.
        else:
            count = 0
            x_input, file_path = [], []
            while count != class_self.sequence_length:
                x, f = get_input(class_self, jpg_paths, pcd_paths)
                if x is not None and f is not None:
                    x_input.append(x)
                    file_path.append(f)
                    count += 1
            x_input = np.array(x_input)
            file_path = np.array(file_path)

        # Set the output.
        y_output = targets

        # Got a proper sample.
        if x_input is not None and y_output is not None and file_path is not None:
            x_inputs.append(x_input)
            y_outputs.append(y_output)
            file_paths.append(file_path)

        assert len(x_inputs) == len(y_outputs)
        assert len(y_outputs) == len(file_paths)

        if verbose is True:
            bar.update(len(x_inputs))

    if verbose is True:
        bar.finish()

    assert len(x_inputs) == size
    assert len(y_outputs) == size

    # Turn everything into ndarrays.
    x_inputs = np.array(x_inputs)
    y_outputs = np.array(y_outputs)

    # Prepare result values.
    assert len(x_inputs) == size
    assert len(y_outputs) == size
    if yield_file_paths is False:
        return_values = (x_inputs, y_outputs)
    else:
        return_values = (x_inputs, y_outputs, file_paths)

    # This is used in multiprocessing. Creates a pickle file and puts the data there.
    if output_queue is not None:
        output_path = uuid.uuid4().hex + ".p"
        pickle.dump(return_values, open(output_path, "wb"))
        output_queue.put(output_path)
    else:
        return return_values


def get_input(class_self, jpg_paths, pcd_paths):
    # Get a random image.
    if class_self.input_type == "image":
        if len(jpg_paths) == 0:
            logger.info("777")
            return None, None
        jpg_path = random.choice(jpg_paths)
        image = class_self._load_image(jpg_path)
        file_path = jpg_path
        x_input = image

    # Get a random voxelgrid.
    elif class_self.input_type == "voxelgrid":
        if len(pcd_paths) == 0:
            return None, None
        pcd_path = random.choice(pcd_paths)
        try:
            voxelgrid = class_self._load_voxelgrid(pcd_path)
            file_path = pcd_path
            x_input = voxelgrid
        except Exception as e:
            print(e)
            return None, None

    # Get a random pointcloud.
    elif class_self.input_type == "pointcloud":
        if len(pcd_paths) == 0:
            return None, None
        pcd_path = random.choice(pcd_paths)
        try:
            pointcloud = class_self._load_pointcloud(pcd_path)
            file_path = pcd_path
            x_input = pointcloud
        except Exception as e:
            print(e)
            return None, None

    # Should not happen.
    else:
        raise Exception("Unknown input_type: " + class_self.input_type)

    return x_input, file_path



from cgmml.common.model_utils.model_utils_constants import BLACKLIST_QRCODES


def preprocess_depthmap(depthmap):
    return depthmap.astype("float32")


def preprocess_targets(targets, targets_indices):
    if targets_indices is not None:
        targets = targets[targets_indices]
    return targets.astype("float32")


def filter_blacklisted_qrcodes(qrcode_paths):
    qrcode_paths_filtered = []
    assert len(qrcode_paths) != 0, 'The provided qrcode_path is empty'
    for qrcode_path in qrcode_paths:
        qrcode_str = qrcode_path.split('/')[-1]
        assert '-' in qrcode_str and len(qrcode_str) == 21, qrcode_str
        if qrcode_str in BLACKLIST_QRCODES:
            continue
        qrcode_paths_filtered.append(qrcode_path)
    return qrcode_paths_filtered



import logging
from pathlib import Path
from typing import Union

import numpy as np

from depthmap import Depthmap

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)


def export_obj(fpath: Union[str, Path],
               dmap: Depthmap,
               floor_altitude_in_meters: float,
               triangulate: bool):
    """Export .obj file, which can be visualized in tools like Meshlab.

    floor_altitude_in_meters is the floor altitude to align floor to Y=zero
    triangulate=True generates OBJ of type mesh
    triangulate=False generates OBJ of type pointcloud
    """
    fpath = Path(fpath)
    count = 0
    indices = np.zeros((dmap.width, dmap.height))

    # Create MTL file (a standart extension of OBJ files to define geometry materials and textures)
    material_fpath = fpath.with_suffix('.mtl')
    if dmap.has_rgb:
        with open(material_fpath, 'w') as f:
            f.write('newmtl default\n')
            f.write(f'map_Kd {str(dmap.rgb_fpath.absolute())}\n')

    with open(fpath, 'w') as f:
        if dmap.has_rgb:
            f.write(f'mtllib {material_fpath.name}\n')
            f.write('usemtl default\n')

        points_3d_arr = dmap.convert_2d_to_3d_oriented()
        for x in range(2, dmap.width - 2):
            for y in range(2, dmap.height - 2):
                depth = dmap.depthmap_arr[x, y]
                if not depth:
                    continue
                count = count + 1
                indices[x, y] = count  # add index of written vertex into array

                x_coord = points_3d_arr[0, x, y]
                y_coord = points_3d_arr[1, x, y] - floor_altitude_in_meters
                z_coord = points_3d_arr[2, x, y]
                f.write(f'v {x_coord} {y_coord} {z_coord}\n')
                f.write(f'vt {x / dmap.width} {y / dmap.height}\n')

        if triangulate:
            _do_triangulation(dmap, indices, f)
        logger.info('Mesh exported into %s', fpath)


def export_renderable_obj(fpath: Union[str, Path],
                          dmap: Depthmap,
                          floor_altitude_in_meters: float,
                          point_size_in_meters: float):
    """Export pointcloud as .obj file, which can be rendered in tools like Blender.

    floor_altitude_in_meters is the floor altitude to align floor to Y=zero
    point_size_in_meters is point size in meters
    """
    fpath = Path(fpath)
    count = 1

    # Create MTL file (a standart extension of OBJ files to define geometry materials and textures)
    material_fpath = fpath.with_suffix('.mtl')
    if dmap.has_rgb:
        with open(material_fpath, 'w') as f:
            f.write('newmtl default\n')
            f.write(f'map_Kd {str(dmap.rgb_fpath.absolute())}\n')

    with open(fpath, 'w') as f:
        if dmap.has_rgb:
            f.write(f'mtllib {material_fpath.name}\n')
            f.write('usemtl default\n')

        points_3d_arr = dmap.convert_2d_to_3d_oriented()
        for x in range(2, dmap.width - 2):
            for y in range(2, dmap.height - 2):
                depth = dmap.depthmap_arr[x, y]
                if not depth:
                    continue

                x_coord = points_3d_arr[0, x, y]
                y_coord = points_3d_arr[1, x, y] - floor_altitude_in_meters
                z_coord = points_3d_arr[2, x, y]
                _write_obj_cube(f, dmap, x, y, count, x_coord, y_coord, z_coord, point_size_in_meters)
                count = count + 8

        logger.info('Mesh exported into %s', fpath)


def _do_triangulation(dmap: Depthmap, indices, filehandle):
    max_diff = 0.2
    for x in range(2, dmap.width - 2):
        for y in range(2, dmap.height - 2):
            # get depth of all points of 2 potential triangles
            d00 = dmap.depthmap_arr[x, y]
            d10 = dmap.depthmap_arr[x + 1, y]
            d01 = dmap.depthmap_arr[x, y + 1]
            d11 = dmap.depthmap_arr[x + 1, y + 1]

            # check if first triangle points have existing indices
            if indices[x, y] > 0 and indices[x + 1, y] > 0 and indices[x, y + 1] > 0:
                # check if the triangle size is valid (to prevent generating triangle
                # connecting child and background)
                if abs(d00 - d10) + abs(d00 - d01) + abs(d10 - d01) < max_diff:
                    c = str(int(indices[x, y]))
                    b = str(int(indices[x + 1, y]))
                    a = str(int(indices[x, y + 1]))
                    # define triangle indices in (world coordinates / texture coordinates)
                    _write_obj_triangle_indices(filehandle, a, b, c)

            # check if second triangle points have existing indices
            if indices[x + 1, y + 1] > 0 and indices[x + 1, y] > 0 and indices[x, y + 1] > 0:
                # check if the triangle size is valid (to prevent generating triangle
                # connecting child and background)
                if abs(d11 - d10) + abs(d11 - d01) + abs(d10 - d01) < max_diff:
                    a = str(int(indices[x + 1, y + 1]))
                    b = str(int(indices[x + 1, y]))
                    c = str(int(indices[x, y + 1]))
                    # define triangle indices in (world coordinates / texture coordinates)
                    _write_obj_triangle_indices(filehandle, a, b, c)


def _write_obj_cube(
        f,
        dmap: Depthmap,
        x: int,
        y: int,
        count: int,
        x_coord: float,
        y_coord: float,
        z_coord: float,
        size: float):

    # cube points
    f.write(f'v {x_coord - size} {y_coord - size} {z_coord - size}\n')
    f.write(f'vt {x / dmap.width} {y / dmap.height}\n')
    f.write(f'v {x_coord + size} {y_coord - size} {z_coord - size}\n')
    f.write(f'vt {x / dmap.width} {y / dmap.height}\n')
    f.write(f'v {x_coord - size} {y_coord + size} {z_coord - size}\n')
    f.write(f'vt {x / dmap.width} {y / dmap.height}\n')
    f.write(f'v {x_coord + size} {y_coord + size} {z_coord - size}\n')
    f.write(f'vt {x / dmap.width} {y / dmap.height}\n')
    f.write(f'v {x_coord - size} {y_coord - size} {z_coord + size}\n')
    f.write(f'vt {x / dmap.width} {y / dmap.height}\n')
    f.write(f'v {x_coord + size} {y_coord - size} {z_coord + size}\n')
    f.write(f'vt {x / dmap.width} {y / dmap.height}\n')
    f.write(f'v {x_coord - size} {y_coord + size} {z_coord + size}\n')
    f.write(f'vt {x / dmap.width} {y / dmap.height}\n')
    f.write(f'v {x_coord + size} {y_coord + size} {z_coord + size}\n')
    f.write(f'vt {x / dmap.width} {y / dmap.height}\n')

    # front face
    _write_obj_triangle_indices(f, str(count + 2), str(count + 1), str(count + 0))
    _write_obj_triangle_indices(f, str(count + 3), str(count + 2), str(count + 1))

    # back face
    _write_obj_triangle_indices(f, str(count + 4), str(count + 5), str(count + 6))
    _write_obj_triangle_indices(f, str(count + 7), str(count + 6), str(count + 5))

    # left face
    _write_obj_triangle_indices(f, str(count + 2), str(count + 4), str(count + 0))
    _write_obj_triangle_indices(f, str(count + 2), str(count + 4), str(count + 6))

    # right face
    _write_obj_triangle_indices(f, str(count + 1), str(count + 5), str(count + 3))
    _write_obj_triangle_indices(f, str(count + 7), str(count + 5), str(count + 3))

    # top face
    _write_obj_triangle_indices(f, str(count + 6), str(count + 3), str(count + 2))
    _write_obj_triangle_indices(f, str(count + 3), str(count + 6), str(count + 7))

    # bottom face
    _write_obj_triangle_indices(f, str(count + 0), str(count + 1), str(count + 4))
    _write_obj_triangle_indices(f, str(count + 4), str(count + 3), str(count + 1))


def _write_obj_triangle_indices(f, a: str, b: str, c: str):
    f.write(f'f {a}/{a} {b}/{b} {c}/{c}\n')


def _write_pcd_header(filehandle, count):
    filehandle.write('# timestamp 1 1 float 0\n')
    filehandle.write('# .PCD v.7 - Point Cloud Data file format\n')
    filehandle.write('VERSION .7\n')
    filehandle.write('FIELDS x y z c\n')
    filehandle.write('SIZE 4 4 4 4\n')
    filehandle.write('TYPE F F F F\n')
    filehandle.write('COUNT 1 1 1 1\n')
    filehandle.write('WIDTH ' + count + '\n')
    filehandle.write('HEIGHT 1\n')
    filehandle.write('VIEWPOINT 0 0 0 1 0 0 0\n')
    filehandle.write('POINTS ' + count + '\n')
    filehandle.write('DATA ascii\n')


def export_pcd(filename: str, dmap: Depthmap):
    with open(filename, 'w') as f:
        count = str(_get_count(dmap))
        _write_pcd_header(f, count)

        for x in range(2, dmap.width - 2):
            for y in range(2, dmap.height - 2):
                depth = dmap.depthmap_arr[x, y]
                if not depth:
                    continue
                res = dmap.convert_2d_to_3d(1, x, y, depth)
                if not res.any():
                    continue
                confidence = dmap.confidence_arr[x, y]
                f.write(str(-res[0]) + ' ' + str(res[1]) + ' ' + str(res[2]) + ' ' + str(confidence) + '\n')
        logger.info('Pointcloud exported into %s', filename)


def _get_count(dmap: Depthmap) -> int:
    count = 0
    for x in range(2, dmap.width - 2):
        for y in range(2, dmap.height - 2):
            depth = dmap.depthmap_arr[x, y]
            if not depth:
                continue
            res = dmap.convert_2d_to_3d(1, x, y, depth)
            if not res.any():
                continue
            count = count + 1
    return count



#
# Child Growth Monitor - Free Software for Zero Hunger
# Copyright (c) 2019 Tristan Behrens <tristan@ai-guru.de> for Welthungerhilfe
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import matplotlib.pyplot as plt
import numpy as np
import itertools
import datetime
import glob
import os
import logging
try:
    import vtk
except Exception:
    pass
    #print("WARNING! VTK not available. This might limit the functionality.")
from pyntcloud import PyntCloud
import pickle
import random
from tensorflow.python.client import device_lib
import multiprocessing
import uuid
from tqdm import tqdm
import traceback
from PIL import Image
import cv2

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)


def load_pcd_as_ndarray(pcd_path):
    """
    Loads a PCD-file. Yields a numpy-array.
    """
    pointcloud = PyntCloud.from_file(pcd_path)
    values = pointcloud.points.values
    return values


def subsample_pointcloud(pointcloud, target_size, subsampling_method="random", dimensions=[0, 1, 2]):
    """
    Yields a subsampled pointcloud.

    These subsamplinge modes are available:
    - "random": Yields a random subset. Multiple occurrences of a single point are possible.
    - "first": Yields the first n points
    - "sequential_skip": Attempts to keep the order of the points intact,
                         might skip some elements if the pointcloud is too big. E.g. every second point is skipped.

    Note: All methods ensure that the target_size is met. If necessary zeroes are appended.
    """

    # Check if the requested subsampling method is all right.
    possible_subsampling_methods = ["random", "first", "sequential_skip"]
    assert subsampling_method in possible_subsampling_methods, "Subsampling method {} not in {}".format(
        subsampling_method, possible_subsampling_methods)

    # Random subsampling.
    if subsampling_method == "random":
        indices = np.arange(0, pointcloud.shape[0])
        indices = np.random.choice(indices, target_size)
        result = pointcloud[indices]

    elif subsampling_method == "first":
        result = np.zeros((target_size, pointcloud.shape[1]), dtype="float32")
        result[:len(pointcloud), :] = pointcloud[:target_size]

    elif subsampling_method == "sequential_skip":
        result = np.zeros((target_size, pointcloud.shape[1]), dtype="float32")
        skip = max(1, round(len(pointcloud) / target_size))
        pointcloud_skipped = pointcloud[::skip, :]
        result = np.zeros((target_size, pointcloud.shape[1]), dtype="float32")
        result[:len(pointcloud_skipped), :] = pointcloud_skipped[:target_size]

    return result[:, dimensions]


def load_vtk(vtk_path):
    """
    Loads a VTK-file. Yields a numpy-array.
    """

    reader = vtk.vtkDataSetReader()
    reader.SetFileName(vtk_path)
    reader.ReadAllVectorsOn()
    reader.ReadAllScalarsOn()
    reader.Update()
    data = reader.GetOutput()

    points = np.zeros((data.GetNumberOfPoints(), 3))

    for i in range(data.GetNumberOfPoints()):
        points[i] = data.GetPoint(i)

    return points


def render_pointcloud(points, title=None):
    """
    Renders a point-cloud.
    """

    fig = plt.figure(figsize=(10, 10))
    ax = fig.add_subplot(111, projection='3d')

    ax.scatter(points[:, 0], points[:, 1], points[:, 2], s=0.5, cmap="gray", alpha=0.5)

    ax.set_xlabel("x")
    ax.set_ylabel("y")
    ax.set_zlabel("z")

    if title is not None:
        plt.title(title)

    plt.show()
    plt.close()


def ensure_voxelgrid_shape(voxelgrid, voxelgrid_target_shape):
    voxelgrid = pad_voxelgrid(voxelgrid, voxelgrid_target_shape)
    voxelgrid = crop_voxelgrid(voxelgrid, voxelgrid_target_shape)
    return voxelgrid


def pad_voxelgrid(voxelgrid, voxelgrid_target_shape):

    pad_before = [0.0] * 3
    pad_after = [0.0] * 3
    for i in range(3):
        pad_before[i] = (voxelgrid_target_shape[i] - voxelgrid.shape[i]) // 2
        pad_before[i] = max(0, pad_before[i])
        pad_after[i] = voxelgrid_target_shape[i] - pad_before[i] - voxelgrid.shape[i]
        pad_after[i] = max(0, pad_after[i])
    voxelgrid = np.pad(
        voxelgrid,
        [(pad_before[0], pad_after[0]), (pad_before[1], pad_after[1]), (pad_before[2], pad_after[2])],
        'constant', constant_values=[(0, 0), (0, 0), (0, 0)]
    )

    return voxelgrid


def crop_voxelgrid(voxelgrid, voxelgrid_target_shape):

    while voxelgrid.shape[0] > voxelgrid_target_shape[0]:
        voxels_start = np.count_nonzero(voxelgrid[0, :, :] != 0.0)
        voxels_end = np.count_nonzero(voxelgrid[-1, :, :] != 0.0)
        if voxels_start > voxels_end:
            voxelgrid = voxelgrid[:-1, :, :]
        else:
            voxelgrid = voxelgrid[1:, :, :]

    while voxelgrid.shape[1] > voxelgrid_target_shape[1]:
        voxels_start = np.count_nonzero(voxelgrid[:, 0, :] != 0.0)
        voxels_end = np.count_nonzero(voxelgrid[:, -1, :] != 0.0)
        if voxels_start > voxels_end:
            voxelgrid = voxelgrid[:, :-1, :]
        else:
            voxelgrid = voxelgrid[:, 1:, :]

    while voxelgrid.shape[2] > voxelgrid_target_shape[2]:
        voxels_start = np.count_nonzero(voxelgrid[:, :, 0] != 0.0)
        voxels_end = np.count_nonzero(voxelgrid[:, :, -1] != 0.0)
        if voxels_start > voxels_end:
            voxelgrid = voxelgrid[:, :, :-1]
        else:
            voxelgrid = voxelgrid[:, :, 1:]

    return voxelgrid


def center_crop_voxelgrid(voxelgrid, voxelgrid_target_shape):

    # Center crop.
    crop_start = [0.0] * 3
    crop_end = [0.0] * 3
    for i in range(3):
        crop_start[i] = (voxelgrid.shape[i] - voxelgrid_target_shape[i]) // 2
        crop_start[i] = max(0, crop_start[i])
        crop_end[i] = voxelgrid_target_shape[i] + crop_start[i]
    voxelgrid = voxelgrid[crop_start[0]:crop_end[0],
                          crop_start[1]:crop_end[1], crop_start[2]:crop_end[2]]

    return voxelgrid


def render_voxelgrid(voxelgrid, title=None):
    """
    Renders a voxel-grid.
    """

    figsize = (5, 5)
    fig = plt.figure(figsize=figsize)
    ax = fig.gca(projection='3d')
    transformed_voxelgrid = np.flip(np.flip(voxelgrid, axis=2), axis=0)

    facecolors = np.zeros(transformed_voxelgrid.shape + (3,))
    prod = itertools.product(range(transformed_voxelgrid.shape[0]),
                             range(transformed_voxelgrid.shape[1]),
                             range(transformed_voxelgrid.shape[2]))
    for x, y, z in prod:
        color = (1.0 - y / 32)
        facecolors[x, y, z, 0] = color
        facecolors[x, y, z, 1] = color
        facecolors[x, y, z, 2] = color

    ax.voxels(transformed_voxelgrid, facecolors=facecolors, edgecolor="k")

    if title is not None:
        plt.title(title)

    plt.show()
    plt.close()


def get_datetime_string():
    """
    Returns a datetime string.
    """

    return datetime.datetime.now().strftime("%Y%m%d-%H%M")


def get_latest_preprocessed_dataset(path=".", filter=""):
    """
    Retrieves the path of the latest preprocessed dataset. Takes into account a filter.
    """
    glob_search_path = os.path.join(path, "*.p")
    paths = [x for x in glob.glob(glob_search_path) if filter in x]
    if len(paths) == 0:
        raise Exception("No datasets found for filter " + filter
                        + " at path " + os.path.abspath(path))
    return sorted(paths)[-1]


def get_latest_model(path=".", filter=""):
    """
    Retrieves the path of the latest preprocessed dataset. Takes into account a filter.
    """
    glob_search_path = os.path.join(path, "*.h5")
    paths = [x for x in glob.glob(glob_search_path) if filter in x]
    if len(paths) == 0:
        raise Exception("No models found for filter " + filter
                        + " at path " + os.path.abspath(path))
    return sorted(paths)[-1]


def pointcloud_to_rgb_map(original_pointcloud, target_width=512, target_height=512, scale_factor=1.5, axis="vertical"):
    """
    Maps a pointcloud to a RGB-image. Stores height, density and intensity as separate channels.
    """
    if axis == "horizontal":

        # Transform to pixel-space.
        scale = np.array([target_width / scale_factor, target_width / scale_factor,
                          target_width / scale_factor, target_width / scale_factor])  # TODO is this okay?
        translate = np.array([target_width / 2, target_height / 2, 0.0, 0.0])
        pointcloud = original_pointcloud * scale + translate

        # Crop the pointcloud.
        crop_mask = np.where(
            (pointcloud[:, 0] >= 0)
            & (pointcloud[:, 0] < target_width)
            & (pointcloud[:, 1] >= 0)
            & (pointcloud[:, 1] < target_height))
        pointcloud = pointcloud[crop_mask]

        # Get indices and counts.
        _, indices, counts = np.unique(
            pointcloud[:, 0:2], axis=0, return_index=True, return_counts=True)

        # Get unique pixel coordinates.
        pixel_coordinates = np.int_(np.array([[x, y] for x, y, _, _ in pointcloud[indices]]))

        # Create the height map.
        heights = pointcloud[indices][:, 2]
        height_map = np.zeros((target_width, target_height))
        height_map[pixel_coordinates[:, 0], pixel_coordinates[:, 1]] = heights
        height_map /= target_width

        # Create the density map.
        densities = np.minimum(1.0, np.log(counts + 1) / np.log(64))
        density_map = np.zeros((target_width, target_height))
        density_map[pixel_coordinates[:, 0], pixel_coordinates[:, 1]] = densities

        # Create the intensity map.
        intensities = pointcloud[indices][:, 3]
        intensity_map = np.zeros((target_width, target_height))
        intensity_map[pixel_coordinates[:, 0], pixel_coordinates[:, 1]] = intensities
        intensity_map /= target_width

        # Compose the RGB-map.
        rgb_map = np.zeros((target_width, target_height, 3))
        rgb_map[:, :, 0] = height_map
        rgb_map[:, :, 1] = density_map
        rgb_map[:, :, 2] = intensity_map

        return rgb_map

    elif axis == "vertical":

        # Transform to pixel-space.
        scale = np.array([target_width / scale_factor, target_width / scale_factor,
                          target_width / scale_factor, target_width / scale_factor])
        # TODO is this okay?
        translate = np.array([target_height / 2, target_width / 2, -target_width / 3, 0.0])
        pointcloud = original_pointcloud * scale + translate

        # Crop the pointcloud.
        crop_mask = np.where(
            (pointcloud[:, 1] >= 0)
            & (pointcloud[:, 1] < target_width)

            & (pointcloud[:, 2] >= 0)
            & (pointcloud[:, 2] < target_height)
        )
        pointcloud = pointcloud[crop_mask]

        # Get indices and counts.
        _, indices, counts = np.unique(
            pointcloud[:, [1, 2]], axis=0, return_index=True, return_counts=True)

        # Get unique pixel coordinates.
        pixel_coordinates = np.int_(np.array([[-x, y] for _, y, x, _ in pointcloud[indices]]))

        # Create the height map.
        heights = pointcloud[indices][:, 0]
        height_map = np.zeros((target_width, target_height))
        height_map[pixel_coordinates[:, 0], pixel_coordinates[:, 1]] = heights
        height_map /= target_width

        # Create the density map.
        densities = np.minimum(1.0, np.log(counts + 1) / np.log(64))
        density_map = np.zeros((target_width, target_height))
        density_map[pixel_coordinates[:, 0], pixel_coordinates[:, 1]] = densities

        # Create the intensity map.
        intensities = pointcloud[indices][:, 3]
        intensity_map = np.zeros((target_width, target_height))
        intensity_map[pixel_coordinates[:, 0], pixel_coordinates[:, 1]] = intensities
        intensity_map /= target_width

        # Compose the RGB-map.
        rgb_map = np.zeros((target_width, target_height, 3))
        rgb_map[:, :, 0] = height_map
        rgb_map[:, :, 1] = density_map
        rgb_map[:, :, 2] = intensity_map

        return rgb_map

    else:
        raise Exception("Unknown axis: " + axis)


def show_rgb_map(rgb_map):
    """
    Renders a RGB-map.
    """

    plt.figure(figsize=(10, 10))
    plt.subplots_adjust(wspace=0.1, hspace=0.1)
    plt.subplot(2, 2, 1)
    show_rgb_map_channel(rgb_map[::-1, :, :], "RGB")
    plt.subplot(2, 2, 2)
    show_rgb_map_channel(rgb_map[::-1, :, 0], "Height", cmap="gray")
    plt.subplot(2, 2, 3)
    show_rgb_map_channel(rgb_map[::-1, :, 1], "Density", cmap="gray")
    plt.subplot(2, 2, 4)
    show_rgb_map_channel(rgb_map[::-1, :, 2], "Intensity", cmap="gray")
    plt.show()
    plt.close()


def show_rgb_map_channel(data, title, cmap=None):
    """
    Renders a channel of a RGB-map.
    """

    fig = plt.imshow(data, cmap=cmap)
    fig.axes.get_xaxis().set_visible(False)
    fig.axes.get_yaxis().set_visible(False)
    plt.title(title)


def find_timestamps_of_trained_models(root_path):
    """
    Extracts the timestamps. Different timestamps will represent different models/trainings.
    """
    all_paths = find_all_history_paths(root_path)
    date_times = []
    for path in all_paths:
        split = path.split("/")[-1].split("-")
        date = split[0]
        time = split[1]
        date_time = date + "-" + time
        date_times.append(date_time)
    date_times = sorted(list(set(date_times)))
    return date_times


def plot_date_times(date_times, all_history_paths, start_index, end_index=100090, key_suffix=None):
    for date_time in date_times:

        # Load all histories for date-time.
        history_paths = [
            history_path for history_path in all_history_paths if date_time in history_path]
        histories = []
        for history_path in history_paths:
            history = pickle.load(open(history_path, "rb"))
            histories.append(history)

        # Plot the histories.
        for history, history_path in zip(histories, history_paths):
            split = history_path.split("/")[-1].split("-")
            for key in history.keys():
                if key_suffix is not None and key_suffix in key:
                    plt.plot(history[key][start_index:end_index],
                             label=key + " " + split[2] + " " + date_time)
    plt.legend()
    plt.show()
    plt.close()


def get_mean_error(date_times, all_history_paths, start_index, end_index=100090, key_suffix=None):
    for date_time in date_times:

        # Load all histories for date-time.
        history_paths = [
            history_path for history_path in all_history_paths if date_time in history_path]
        histories = []
        for history_path in history_paths:
            history = pickle.load(open(history_path, "rb"))
            histories.append(history)

        # Print the average error of each model.
        for history, history_path in zip(histories, history_paths):
            split = history_path.split("/")[-1].split("-")
            for key in history.keys():
                if key_suffix is not None and key_suffix in key:
                    lst = history[key][start_index:end_index]
                    avg_error = sum(lst) / len(lst)
                    logger.info('avg %d %d %d between epoch %s and %s = %s', key, split[2], date_time,
                                str(start_index), str(end_index), str(avg_error))


def find_all_history_paths(root_path):
    all_paths = glob.glob(os.path.join(root_path, "*.p"))
    history_paths = [path for path in all_paths if "history" in path]
    return history_paths


def create_training_tasks(qrcodes, subset_sizes, random_seed=666):
    """
    Takes a bunch of qr-codes and creates subsets of given sizes.
    """
    random_seed = 666
    randomizer = random.Random(random_seed)
    qrcodes_tasks = []
    for subset_size in subset_sizes:
        randomizer.shuffle(qrcodes)
        qrcodes_shuffle_subset = qrcodes[:int(len(qrcodes) * subset_size)]
        split_index = int(0.8 * len(qrcodes_shuffle_subset))
        qrcodes_train = sorted(qrcodes_shuffle_subset[:split_index])
        qrcodes_validate = sorted(qrcodes_shuffle_subset[split_index:])
        qrcodes_tasks.append((qrcodes_train, qrcodes_validate))
    return qrcodes_tasks


def get_available_gpus():
    local_device_protos = device_lib.list_local_devices()
    return [x.name for x in local_device_protos if x.device_type == 'GPU']


def multiprocess(
    entries,
    process_method,
    number_of_workers=None,
    process_individial_entries=True,
    pass_process_index=False,
    progressbar=False,
    disable_gpu=False
):
    # Get number of workers.
    if number_of_workers is None:
        number_of_workers = multiprocessing.cpu_count()
    logger.info("Using %d workers.", number_of_workers)
    if disable_gpu is True:
        logger.info("GPU is disabled.")

    # Split into list.
    entry_sublists = np.array_split(entries, number_of_workers)
    assert len(entry_sublists) == number_of_workers
    assert np.sum([len(entry_sublist) for entry_sublist in entry_sublists]) == len(entries)

    # Define an output queue
    output = multiprocessing.Queue()

    # This method is starting point for multiprocessing
    def process_entries(entry_sublist, process_index):

        if disable_gpu is True:
            os.environ["CUDA_VISIBLE_DEVICES"] = "-1"

        try:
            # Process individually.
            if process_individial_entries:
                for entry_index, entry in enumerate(tqdm(entry_sublist, position=process_index)):
                    if pass_process_index is False:
                        result = process_method(entry)
                    else:
                        result = process_method(entry, process_index)

            # Process all.
            else:
                if pass_process_index is False:
                    result = process_method(entry_sublist)
                else:
                    result = process_method(entry_sublist, process_index)

            # No results returned. Just provide status string.
            if result is None:
                output.put("Processed {}".format(len(entry_sublist)))

            # Results found. Playing it safe. Use pickle.
            else:
                pickle_path = str(uuid.uuid4()) + ".pickletemp"
                pickle.dump(result, open(pickle_path, "wb"))
                output.put(pickle_path)

        # Handle keyboard interrupt.
        except KeyboardInterrupt:
            output.put("Keyboard interrupt.")

        # TODO remove later
        except Exception as e:
            print("\n\n\n\n\n\n\n\n\n\n\n")
            print(e)
            traceback.print_exc()

    # Setup a list of processes that we want to run
    processes = [multiprocessing.Process(target=process_entries, args=(
        entry_sublist, process_index)) for process_index, entry_sublist in enumerate(entry_sublists)]

    # Run processes
    for process in processes:
        process.start()

    # Exit the completed processes
    try:
        for process in processes:
            process.join()
    except KeyboardInterrupt:
        print("Keyboard interrupt. Gracefully terminating multi-processing...")
        for process in processes:
            process.terminate()
            process.join()
        return

    # Get process results from the output queue.
    results = []
    for _ in processes:
        result = output.get()

        # Results are in a pickle file.
        if result.endswith(".pickletemp"):
            results.append(pickle.load(open(result, "rb")))
            os.remove(result)

        # Just plain results.
        else:
            results.append(result)
    return results


# Render a subsample of the artifacts.
def render_artifacts_as_gallery(artifacts, targets=None, qr_code=None, timestamp=None,
                                num_columns=10, target_size=(1920 // 4, 1080 // 4), image_path=None, use_plt=True):

    # Render results image.
    result_images = []
    row_images = []
    for artifact in artifacts:
        if isinstance(artifact, str) is False:
            artifact = artifact[0]
        path = artifact.replace("whhdata", "localssd")
        img = Image.open(path)
        img = img.resize(target_size)
        img = np.array(img)
        img = np.rot90(img, 3)
        row_images.append(img)
        if len(row_images) == num_columns:
            row_image = np.hstack(row_images)
            result_images.append(row_image)
            row_images = []
    # Handle last row.
    if len(row_images) != 0:
        while len(row_images) != num_columns:
            black_image = np.zeros((target_size) + (3,)).astype("uint8")
            row_images.append(black_image)
        row_image = np.hstack(row_images)
        result_images.append(row_image)
    result_image = np.vstack(result_images)

    # Create title string.
    title_string = ""
    if qr_code is not None:
        title_string += "QR-code: " + qr_code
    if timestamp is not None:
        title_string += "Timestamp: " + str(timestamp)
    if targets is not None:
        title_string += " Targets: " + ", ".join([str(target) for target in targets])

    # Render with plt.
    if use_plt is True:
        plt.figure(figsize=(20, int(20 * result_image.shape[0] / result_image.shape[1])))
        plt.imshow(result_image)
        plt.axis("off")
        plt.title(title_string)

        if image_path is None:
            plt.show()
        else:
            plt.savefig(image_path)
        plt.close()

    else:
        # Write image to hard drive.
        cv2.imwrite(image_path, result_image[:, :, ::-1])



import logging
import time
from typing import List

import numpy as np
import tensorflow as tf
from tensorflow.keras.models import load_model

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)


def get_prediction_uncertainty_deepensemble(model_paths: list,
                                            dataset_evaluation: tf.data.Dataset,
                                            batch_size: int = 256) -> np.ndarray:
    """Predict standard deviation of multiple predictions with different dropouts
    Args:
        model_path: Path of the trained model
        dataset_evaluation: dataset in which the evaluation need to performed
    Returns:
        predictions, array shape (N_SAMPLES, )
    """
    dataset = dataset_evaluation.batch(batch_size)
    logger.info("Start predicting uncertainty")

    # Go through all models and compute standard deviation of predictions
    start = time.time()
    predictions_list = [_load_and_predict(model_path, dataset) for model_path in model_paths]
    end = time.time()
    logger.info("Total time for uncertainty prediction experiment: %f:.3 sec", end - start)

    std = _calculate_std(predictions_list)
    return std


def _load_and_predict(model_path, dataset):
    logger.info("Loading model from %s", model_path)
    model = load_model(model_path, compile=False)
    logger.info("Predicting with model %s", model_path)
    return _predict(model, dataset)


def _predict(model: tf.Module, dataset: tf.data.Dataset) -> np.ndarray:
    predictions_batches = []
    for x, _y in dataset.as_numpy_iterator():
        prediction_batch = model.predict(x)  # shape (BATCH_SIZE, 1)
        predictions_batches.append(prediction_batch)
    predictions = np.concatenate(predictions_batches)
    return predictions


def _calculate_std(predictions_per_model: List[np.array]) -> np.ndarray:
    predictions_per_model_ = np.array(predictions_per_model)
    std = np.std(predictions_per_model_, axis=0)
    return std



import os
import shutil

import sys
import logging

import pcd2depth
from depthmap import parse_calibration

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)


if __name__ == "__main__":
    if len(sys.argv) != 3:
        print('You did not enter pcd_dir folder and calibration file path')
        print('E.g.: python convertpcd2depth.py pcd_dir calibration_file')
        sys.exit(1)

    pcd_dir = sys.argv[1]
    calibration_file = sys.argv[2]

    calibration = parse_calibration(calibration_file)

    depth_filenames = []
    for (dirpath, dirnames, filenames) in os.walk(pcd_dir):
        depth_filenames.extend(filenames)
    depth_filenames.sort()
    try:
        shutil.rmtree('output')
    except BaseException:
        print('no previous data to delete')
    os.makedirs('output/depth')

    # works for lenovo
    width = int(240 * 0.75)
    height = int(180 * 0.75)

    for filename in depth_filenames:
        input_filename = f'{pcd_dir}/{filename}'
        depthmap = pcd2depth.process(calibration, input_filename, width, height)
        output_filename = f'output/depth/{filename}.depth'
        pcd2depth.write_depthmap(output_filename, depthmap, width, height)
    logger.info('Data exported into folder output')



from decimal import Decimal as D  # noqa: N817


class LmsBoxCox:
    def __init__(self, chart: str, weight: str, muac: str,
                 age_in_days: str, sex: str, height: str):
        self.chart = chart
        self.weight = weight
        self.muac = muac
        self.age_in_days = age_in_days
        self.sex = sex
        self.height = height

    def get_file_name_from_data(self):

        if self.chart == 'wfl' and D(self.height) > 110:
            table_chart = 'wfh'
            table_age = '2_5'
        elif self.chart == 'wfh' and D(self.height) < 65:
            table_chart = 'wfl'
            table_age = '0_2'
        else:
            table_chart = self.chart
            if self.chart == 'wfl':
                table_age = '0_2'
            if self.chart == 'wfh':
                table_age = '2_5'

        if self.sex == 'M':
            table_sex = 'boys'
        elif self.sex == 'F':
            table_sex = 'girls'

        if self.chart in ["wfa", "lhfa"]:
            table_age = "0_5"
            table_chart = self.chart

        self.table = "%(table_chart)s_%(table_sex)s_%(table_age)s" %\
            {"table_chart": table_chart,
             "table_sex": table_sex,
             "table_age": table_age}

    def get_lms_value(self, growth: dict):
        table_data = growth[self.table]
        if self.chart in ["wfh", "wfl"]:
            if D(self.height) < 45:
                raise Exception("too short")
            if D(self.height) > 120:
                raise TypeError("too tall")
            closest_height = float("{0:.1f}". format(D(self.height)))
            scores = table_data.get(str(closest_height))
            if scores is not None:
                self.scores = scores
            else:
                raise TypeError("Score not found for height :%s", (closest_height))

        if self.chart in ['wfa', 'lhfa']:
            scores = table_data.get(str(self.age_in_days))
            if scores is not None:
                self.scores = scores
            else:
                raise TypeError("Scores not found by Day: %s", self.age_in_days)

    def resolve_lms_value(self) -> int:
        skew = D(self.scores.get("L"))
        median = D(self.scores.get("M"))
        coff = D(self.scores.get("S"))
        if self.chart in ['wfa', 'wfl', 'wfh']:
            measurement = D(self.weight)
        elif self.chart == 'lhfa':
            measurement = D(self.height)
        else:
            raise NameError(f"Unknown chart name: {self.chart}")
        return skew, median, coff, measurement



import pandas as pd
import logging

from glob2 import glob
from pathlib import Path

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)

ACCURACY_THRESHOLD = 2
CSV_PATH = "./outputs/**/inaccurate_scans_*.csv"
REPORT_CSV = 'inaccurate_scan_report.csv'


def merge_qrc(row):
    """
    Function to combine qrcodes with scantypes
    """
    qrcode = row['qrcode']
    scantype = row['scantype']
    scans = f"{qrcode}_{scantype}"
    return scans


def filter_scans(dataframe: pd.DataFrame, accuracy: int) -> pd.DataFrame:
    """
    Function that filter dataframe for the fiven accuracy number
    """
    error = dataframe[dataframe['error'].abs() >= accuracy]
    return error


def frame_to_set(dataframe: pd.DataFrame) -> set:
    """
    Function to convert dataframe column to list
    """
    return set(dataframe['scan_code'].to_list())


def calculate_union(set1: set, set2: set) -> set:
    """
    Function to calculate union of two sets
    """
    union_set = set1.union(set2)
    return union_set


def calculate_intersection(set1: set, set2: set) -> set:
    """
    Function to calculate intersection of two sets
    """
    intersection_set = set1.intersection(set2)
    return intersection_set


def extract_model_name(path_name) -> str:
    """
    Function to extract the model name from the path.
    """
    assert path_name.endswith('.csv')
    model_name = Path(path_name).stem
    return model_name


def calculate_inaccurate_scans(csv_filepath: str) -> set:
    """
    Function to combine the models resultant csv files into a single file
    """
    assert csv_filepath.endswith('.csv')
    result_list = pd.read_csv(csv_filepath)
    grouped_result = result_list.groupby(['qrcode', 'scantype'], as_index=False).mean()
    accuracy_df = filter_scans(grouped_result, ACCURACY_THRESHOLD)
    accuracy_df['scan_code'] = accuracy_df.apply(merge_qrc, axis=1)
    csv_name = csv_filepath.split('/')[-1]
    file_name = f"file_{csv_name}"
    accuracy_df.to_csv(file_name, index=False)
    frame_set = frame_to_set(accuracy_df)
    return frame_set


if __name__ == "__main__":
    csv_files = glob(CSV_PATH)
    if len(csv_files) != 2:
        logger.warning("path contains 0 or more than 2 csv files")

    scan_sets = [calculate_inaccurate_scans(filepath) for filepath in csv_files]
    union_set = calculate_union(scan_sets[0], scan_sets[1])
    inaccurate_scans_intersection = calculate_intersection(scan_sets[0], scan_sets[1])
    inaccurate_scans_intersection_ratio = (len(inaccurate_scans_intersection) / len(union_set)) * 100
    inaccurate_scan_data = [[extract_model_name(csv_files[0]),
                             extract_model_name(csv_files[1]),
                             inaccurate_scans_intersection_ratio,
                             len(union_set),
                             len(inaccurate_scans_intersection)]]
    columns = ['model_1', 'model_2', 'ratio_intersection_over_union',
               'number_of_union_of_inaccurate_scans', 'number_of_common_inaccurate_scans']
    frame = pd.DataFrame(inaccurate_scan_data, columns=columns)
    frame.to_csv(REPORT_CSV)



"""
This is a data-generator that works on the raw data. The raw data is the data that is created by the mobile app.

Note: This file is only here for historical reasons. And it might be gone anytime.
"""

from __future__ import absolute_import
import os
import logging
import numpy as np
import glob2
import json
import random
import keras.preprocessing.image as image_preprocessing
import progressbar
from pyntcloud import PyntCloud
import matplotlib.pyplot as plt
import multiprocessing as mp
import uuid
import pickle
from . import utils

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)


class DataGenerator(object):
    """
    This class generates data for training.
    """

    def __init__(
        self,
        dataset_path,
        input_type,
        output_targets,
        sequence_length=0,
        image_target_shape=(160, 90),
        voxelgrid_target_shape=(32, 32, 32),
        voxel_size_meters=0.01,
        voxelgrid_random_rotation=False,
        pointcloud_target_size=32000,
        pointcloud_random_rotation=False
    ):
        """
        Initializes a DataGenerator.

        Args:
            dataset_path (string): Where the raw data is.
            input_type (string): Specifies how the input-data for the Neural Network looks like.
                                 Either 'image', 'pointcloud', 'voxgrid'.
            output_targets (list of strings): A list of targets for the Neural Network.
                                              For example *['height', 'weight']*.
            sequence_length (int): Specifies the lenght of the sequences. 0 would yield no sequence at all.
            image_target_shape (2D tuple of ints): Target shape of the images.
            voxelgrid_target_shape (3D tuple of ints): Target shape of the voxelgrids.
            voxel_size_meters (float): Size of the voxels. That is, edge length.
            voxelgrid_random_rotation (bool): If True voxelgrids will be rotated randomly.
            pointcloud_target_size (int): Target size of the pointclouds.
            pointcloud_random_rotation (bool): If True pointclouds will be rotated randomly.

        """

        # Preconditions.
        assert os.path.exists(dataset_path), "dataset_path must exist: " + str(dataset_path)
        assert isinstance(input_type, str), "input_type must be string: " + str(input_type)
        assert isinstance(output_targets, list), "output_targets must be list: " + \
            str(output_targets)
        if input_type == "image":
            assert len(image_target_shape) == 2, "image_target_shape must be 2-dimensional: " + \
                str(image_target_shape)
        if input_type == "voxelgrid":
            assert len(voxelgrid_target_shape) == 3, "voxelgrid_target_shape must be 3-dimensional: " + \
                str(voxelgrid_target_shape)

        # Assign the instance-variables.
        self.dataset_path = dataset_path
        self.input_type = input_type
        self.output_targets = output_targets
        self.sequence_length = sequence_length
        self.image_target_shape = image_target_shape
        self.voxelgrid_target_shape = voxelgrid_target_shape
        self.voxel_size_meters = voxel_size_meters
        self.voxelgrid_random_rotation = voxelgrid_random_rotation
        self.pointcloud_target_size = pointcloud_target_size
        self.pointcloud_random_rotation = pointcloud_random_rotation

        # Create some caches.
        self.image_cache = {}
        self.voxelgrid_cache = {}
        self.pointcloud_cache = {}

        # Get all the paths.
        self._get_paths()

        # Check if paths are fine.
        if self.input_type == "image":
            assert self.jpg_paths != []
        elif self.input_type == "voxelgrid" or self.input_type == "pointcloud":
            assert self.pcd_paths != []
        else:
            raise Exception("Unexpected: " + self.input_type)
        assert self.json_paths_personal != []
        assert self.json_paths_measures != []

        # Find all QR-codes.
        self._find_qrcodes()
        assert self.qrcodes != [], "No QR-codes found!"

        # Create the QR-codes dictionary.
        self._create_qrcodes_dictionary()

    def _get_paths(self):
        """
        Retrieves all the relevant paths.

        That is: Paths of JPGs, PCDs, and JSONs.
        """

        logger.info(self.dataset_path)

        # Getting the paths for images.
        glob_search_path = os.path.join(self.dataset_path, "storage/person", "**/*.jpg")
        self.jpg_paths = glob2.glob(glob_search_path)

        # Getting the paths for point clouds.
        glob_search_path = os.path.join(self.dataset_path, "storage/person", "**/*.pcd")
        self.pcd_paths = glob2.glob(glob_search_path)

        # Getting the paths for personal and measurement.
        glob_search_path = os.path.join(self.dataset_path, "**/*.json")
        json_paths = glob2.glob(glob_search_path)
        self.json_paths_personal = [
            json_path for json_path in json_paths if "measures" not in json_path]
        self.json_paths_measures = [
            json_path for json_path in json_paths if "measures" in json_path]
        del json_paths

    def _find_qrcodes(self):
        """
        Finds all QR-codes.

        Each individual is represented via a unique QR-codes. This method extracts the set of QR-codes.
        """

        # Go through all the measures and extract their QR-codes.
        qrcodes = []
        for json_path_measure in self.json_paths_measures:
            json_path_measure_file = open(json_path_measure)
            json_data_measure = json.load(json_path_measure_file)
            qrcode = self._extract_qrcode(json_data_measure)
            qrcodes.append(qrcode)
            json_path_measure_file.close()

        # Provide a sorted set.
        qrcodes = sorted(list(set(qrcodes)))

        self.qrcodes = qrcodes

    def _create_qrcodes_dictionary(self):
        """
        Creates a QR-Code-dictionary.

        This basically sorts all PCDs and JPGs.
        With respect to the targets and the QR-Codes.
        This is used heavily during data generation.
        Takes into account timestamps in order to connect data and measures.
        """

        qrcodes_dictionary = {}

        # Go thorugh all measures.
        for json_path_measure in self.json_paths_measures:

            # Load the data and get type.
            measure_file = open(json_path_measure)
            json_data_measure = json.load(measure_file)
            measure_type = json_data_measure["type"]["value"]
            measure_file.close()

            # Ensure manual data. If it is not a manual measurement, skip.
            if measure_type != "manual":
                continue

            # Extract the QR-code.
            qrcode = self._extract_qrcode(json_data_measure)

            # Create an array in the dictionary if necessary.
            if qrcode not in qrcodes_dictionary.keys():
                qrcodes_dictionary[qrcode] = []

            # Extract the targets from the JSON-data.
            targets = self._extract_targets(json_data_measure)

            # Extract the timestamp from the JSON-data.
            timestamp = self._extract_timestamp_from_path(json_path_measure)

            # Filter paths for qrcodes and measurements. Find all JPGs and PCDs for a
            # given QR-code and make sure that the timestamps are related.
            jpg_paths = [jpg_path for jpg_path in self.jpg_paths if self._is_matching_measurement(
                jpg_path, qrcode, timestamp) is True]
            pcd_paths = [pcd_path for pcd_path in self.pcd_paths if self._is_matching_measurement(
                pcd_path, qrcode, timestamp) is True]

            # Store it all.
            qrcodes_dictionary[qrcode].append((targets, jpg_paths, pcd_paths))

        self.qrcodes_dictionary = qrcodes_dictionary

    def _extract_targets(self, json_data_measure):
        """
        Extracts a list of targets from JSON.
        """

        targets = []
        for output_target in self.output_targets:
            value = json_data_measure[output_target]["value"]
            targets.append(value)
        return targets

    def _extract_qrcode(self, json_data_measure):
        """
        Extracts a QR-code from a JSON.
        """

        person_id = json_data_measure["personId"]["value"]
        json_path_personal = [
            json_path for json_path in self.json_paths_personal if person_id in json_path]
        json_path_personal = [
            json_path for json_path in json_path_personal if "ipynb_checkpoints" not in json_path]
        assert len(json_path_personal) == 1, "Found {} jsons for person_id {}\n{}".format(
            len(json_path_personal), person_id, json_path_personal)
        json_path_personal = json_path_personal[0]
        json_data_personal_file = open(json_path_personal)
        json_data_personal = json.load(json_data_personal_file)
        json_data_personal_file.close()
        qrcode = json_data_personal["qrcode"]["value"]
        return qrcode

    def _is_matching_measurement(self, path, qrcode, timestamp, threshold=(60 * 60 * 24 * 1000)):
        """
        Returns True if timetamps match.

        Given a timestamp it extracts a second one from the path.
        It then computes the difference between those two timestamps.
        If the differences are lower than the threshold it is a match.
        And the QR-code must match too.

        Args:
            path (string): Path to some file.
            qrcode (string): A QR-code that is supposed to be related to the file.
            timestamp (string): A timestamp that is supposed to be related to the file.
            threshold (string): A threshold for a match. In milliseconds. Default is one day.

        Returns:
            type: True if it is a match. False otherwise.
        """

        if qrcode not in path:
            return False

        if "measurements" not in path:
            return False

        # Extract the timestamp from the path. Compute difference. Decide.
        path_timestamp = self._extract_timestamp_from_path(path, qrcode)
        difference = abs(int(timestamp) - int(path_timestamp))
        if difference > threshold:
            return False

        return True

    def _extract_timestamp_from_path(self, file_path, qrcode=None):
        """
        Extracts a timestamp from a path.
        """
        # This is fix of the "underscores in QR-code"-issue. Poor man's solution ;)
        if qrcode is not None:
            file_path = file_path.replace(qrcode, "REPLACED")

        timestamp = file_path.split(os.sep)[-1].split("_")[2]
        assert len(timestamp) == 13, "Invalid timestamp {} encountered for {}".format(
            timestamp, file_path)
        assert timestamp.isdigit()
        return timestamp

    def _load_image(self, image_path):
        """
        Loads an image from a given path.

        Makes use of a cache. Ensures that the loaded images has a target size.
        """

        image = self.image_cache.get(image_path, [])
        if image == []:
            image = image_preprocessing.load_img(image_path, target_size=self.image_target_shape)
            image = image.rotate(-90, expand=True)  # Rotation is necessary.
            image = np.array(image)
            self.voxelgrid_cache[image_path] = image
        return image

    def _load_pointcloud(self, pcd_path, preprocess=True, augmentation=True):

        pointcloud = self.pointcloud_cache.get(pcd_path, [])
        if pointcloud == []:
            pointcloud = PyntCloud.from_file(pcd_path).points.values
            pointcloud = np.array(pointcloud)

            if self.pointcloud_target_size is not None and preprocess is True:
                pointcloud = pointcloud[:self.pointcloud_target_size]
                if len(pointcloud) < self.pointcloud_target_size:
                    zeros = np.zeros((self.pointcloud_target_size - len(pointcloud), 4))
                    pointcloud = np.concatenate([pointcloud, zeros])

            if self.pointcloud_random_rotation is True and augmentation is True:
                numpy_points = pointcloud[:, 0:3]
                numpy_points = self._rotate_point_cloud(numpy_points)
                pointcloud[:, 0:3] = numpy_points

            #self.pointcloud_cache[pcd_path] = pointcloud
        return pointcloud

    def _load_voxelgrid(self, pcd_path, preprocess=True, augmentation=True):
        voxelgrid = self.voxelgrid_cache.get(pcd_path, [])
        if voxelgrid == []:

            # Load the pointcloud.
            point_cloud = PyntCloud.from_file(pcd_path)
            if self.voxelgrid_random_rotation is True and augmentation is True:
                points = point_cloud.points
                numpy_points = points.values[:, 0:3]
                numpy_points = self._rotate_point_cloud(numpy_points)
                points.iloc[:, 0:3] = numpy_points
                point_cloud.points = points

            # Create voxelgrid from pointcloud.
            voxelgrid_id = point_cloud.add_structure(
                "voxelgrid",
                size_x=self.voxel_size_meters,
                size_y=self.voxel_size_meters,
                size_z=self.voxel_size_meters)
            voxelgrid = point_cloud.structures[voxelgrid_id].get_feature_vector(mode="density")

            # Do the preprocessing.
            if preprocess is True:
                voxelgrid = utils.ensure_voxelgrid_shape(voxelgrid, self.voxelgrid_target_shape)
                assert voxelgrid.shape == self.voxelgrid_target_shape

            #self.voxelgrid_cache[pcd_path] = voxelgrid # TODO cache is turned off because of you know why...

        return voxelgrid

    def _rotate_point_cloud(self, point_cloud):

        rotation_angle = np.random.uniform() * 2 * np.pi
        cosval = np.cos(rotation_angle)
        sinval = np.sin(rotation_angle)
        rotation_matrix = np.array([[cosval, sinval, 0],
                                    [-sinval, cosval, 0],
                                    [0, 0, 1]])

        rotated_data = np.zeros(point_cloud.shape, dtype=np.float32)
        for k in range(point_cloud.shape[0]):

            shape_pc = point_cloud[k, ...]
            rotated_data[k, ...] = np.dot(shape_pc.reshape((-1, 3)), rotation_matrix)

        return rotated_data

    def print_statistics(self):

        for qr_code, array in self.qrcodes_dictionary.items():
            logger.info("QR-Code %s has %d different manual measurements", qr_code, len(array))
            for (targets, jpg_paths, pcd_paths) in array:
                logger.info(" Target %s with %d JPGs and %d PCDs.", targets, len(jpg_paths), len(pcd_paths))

        #qrcodes_dictionary[qrcode].append((targets, jpg_paths, pcd_paths))

    def get_input_shape(self):

        if self.input_type == "image":
            return (90, 160, 3)

        elif self.input_type == "voxelgrid":
            return (32, 32, 32)

        elif self.input_type == "pointcloud":
            return (32000, 4)

        else:
            raise Exception("Unknown input_type: " + self.input_type)

    def get_output_size(self):

        return len(self.output_targets)

    def generate(self, size, qrcodes_to_use=None, verbose=False, yield_file_paths=False, multiprocessing_jobs=1):

        if qrcodes_to_use is None:
            qrcodes_to_use = self.qrcodes

        # Main loop.
        while True:

            # Use only a single process.
            if multiprocessing_jobs == 1:
                yield generate_data(self, size, qrcodes_to_use, verbose, yield_file_paths, None)

            # Use multiple processes.
            elif multiprocessing_jobs > 1:

                # Create chunks of almost equal size.
                subset_sizes = [0] * multiprocessing_jobs
                subset_sizes[0:multiprocessing_jobs - 1] = [size
                                                            // multiprocessing_jobs] * (multiprocessing_jobs - 1)
                subset_sizes[multiprocessing_jobs - 1] = size - \
                    sum(subset_sizes[0:multiprocessing_jobs - 1])
                subset_sizes = [s for s in subset_sizes if s > 0]
                assert sum(subset_sizes) == size

                # Create an output_queue.
                output_queue = mp.Queue()

                # Create the processes.
                processes = []
                for subset_size in subset_sizes:
                    process_target = generate_data
                    process_args = (self, subset_size, qrcodes_to_use,
                                    verbose, yield_file_paths, output_queue)
                    process = mp.Process(target=process_target, args=process_args)
                    processes.append(process)

                # Start the processes.
                for p in processes:
                    p.start()

                # Exit the completed processes.
                for p in processes:
                    p.join()

                # Get process results from the output queue
                output_paths = [output_queue.get() for p in processes]

                # Merge data.
                x_inputs_arrays = []
                y_outputs_arrays = []
                file_paths_arrays = []
                for output_path in output_paths:
                    # Read data from file and delete it.
                    result_values = pickle.load(open(output_path, "rb"))
                    assert len(result_values[0]) > 0
                    assert len(result_values[1]) > 0
                    os.remove(output_path)

                    # Gather the data into arrays.
                    if x_inputs_arrays != []:
                        assert result_values[0].shape[1:] == x_inputs_arrays[-1].shape[1:], str(
                            result_values[0].shape) + " vs " + str(x_inputs_arrays[-1].shape)
                    if y_outputs_arrays != []:
                        assert result_values[1].shape[1:] == y_outputs_arrays[-1].shape[1:], str(
                            result_values[1].shape) + " vs " + str(y_outputs_arrays[-1].shape)
                    x_inputs_arrays.append(result_values[0])
                    y_outputs_arrays.append(result_values[1])
                    if yield_file_paths is True:
                        file_paths_arrays.append(result_values[2])
                    else:
                        file_paths_arrays.append([])

                x_inputs = np.concatenate(x_inputs_arrays)
                y_outputs = np.concatenate(y_outputs_arrays)
                file_paths = np.concatenate(file_paths_arrays)
                assert len(x_inputs) == size
                assert len(y_outputs) == size

                # Done.
                if yield_file_paths is False:
                    yield x_inputs, y_outputs
                else:
                    yield x_inputs, y_outputs, file_paths

            else:
                raise Exception("Unexpected value for 'multiprocessing_jobs' "
                                + str(multiprocessing_jobs))

    def generate_dataset(self, qrcodes_to_use=None):

        if qrcodes_to_use is None:
            qrcodes_to_use = self.qrcodes

        x_qrcodes = []
        x_inputs = []
        y_outputs = []
        for index, qrcode in enumerate(qrcodes_to_use):

            logger.info("Processing: %s", qrcode)

            # Get targets and paths.
            if qrcode not in self.qrcodes_dictionary.keys():
                logger.info("No data for: %s", qrcode)
                continue
            targets, jpg_paths, pcd_paths = self.qrcodes_dictionary[qrcode]
            logger.info(targets)

            # Process image.
            if self.input_type == "image":

                for jpg_path in jpg_paths:
                    image = self._load_image(jpg_path)
                    x_qrcodes.append(qrcode)
                    x_inputs.append(image)
                    y_outputs.append(targets)

            # Process voxelgrid.
            elif self.input_type == "voxelgrid":

                for pcd_path in pcd_paths:
                    try:
                        voxelgrid = self._load_voxelgrid(pcd_path)
                    except Exception as e:
                        print(e)
                        print("Error:", pcd_path)

                    x_qrcodes.append(qrcode)
                    x_inputs.append(voxelgrid)
                    y_outputs.append(targets)

            # Process pointcloud.
            elif self.input_type == "pointcloud":

                for pcd_path in pcd_paths:
                    try:
                        pointcloud = self._load_pointcloud(pcd_path)
                    except Exception as e:
                        print(e)
                        print("Error:", pcd_path)
                        continue

                    x_qrcodes.append(qrcode)
                    x_inputs.append(pointcloud)
                    y_outputs.append(targets)

            else:
                raise Exception("Unknown input_type: " + self.input_type)

        x_qrcodes = np.array(x_qrcodes)
        x_inputs = np.array(x_inputs)
        y_outputs = np.array(y_outputs)

        return x_qrcodes, x_inputs, y_outputs

    def analyze_files(self):

        logger.info("Number of JPGs: %d", len(self.jpg_paths))
        logger.info("Number of PCDs: %d", len(self.pcd_paths))
        logger.info("Number of JSONs (personal): %d", len(self.json_paths_personal))
        logger.info("Number of JSONs (measures): %d", len(self.json_paths_measures))

    def analyze_targets(self):
        """
        Extracts and analyzes all targets from the dataset.
        """

        all_targets = []
        for _, targets_array in self.qrcodes_dictionary.items():
            for (targets, _, _) in targets_array:
                all_targets.append(targets)

        x = [targets[0] for targets in all_targets]
        y = [targets[1] for targets in all_targets]
        plt.scatter(x, y)
        plt.xlabel(self.output_targets[0])
        plt.ylabel(self.output_targets[1])
        plt.title("Distribution of " + str(len(all_targets)) + " targets.")
        plt.show()
        plt.close()

    def analyze_pointclouds(self):

        logger.info("Analyzing pointclouds...")
        pointcloud_sizes = []
        bar = progressbar.ProgressBar(max_value=len(self.pcd_paths))
        for index, pcd_path in enumerate(self.pcd_paths):
            try:
                pointcloud = self._load_pointcloud(pcd_path, preprocess=False, augmentation=False)
                pointcloud_sizes.append(pointcloud.shape[0])
            except ValueError:
                pass
            bar.update(index)
        bar.finish()

        logger.info("Rendering histogram...")
        plt.hist(pointcloud_sizes)
        #plt.xlabel(self.output_targets[0])
        #plt.ylabel(self.output_targets[1])
        plt.title("Distribution of pointclouds-sizes.")
        plt.show()
        plt.close()

    def analyze_voxelgrids(self):

        logger.info("Analyzing voxelgrids...")
        voxelgrid_sizes = [0] * len(self.pcd_paths)
        numbers_of_voxels = [0] * len(self.pcd_paths)
        voxel_densities = [0] * len(self.pcd_paths)
        bar = progressbar.ProgressBar(max_value=len(self.pcd_paths))
        for index, pcd_path in enumerate(self.pcd_paths):
            try:
                voxelgrid = self._load_voxelgrid(pcd_path, augmentation=False, preprocess=False)
                voxelgrid_size = voxelgrid.shape[0]
                voxelgrid_sizes[index] = voxelgrid_size
                number_of_voxels = np.count_nonzero(voxelgrid != 0.0)
                numbers_of_voxels[index] = number_of_voxels
                voxel_densities[index] = number_of_voxels / (voxelgrid_size ** 3)
            except ValueError as error:
                print(pcd_path)
                print(error)
            bar.update(index)
        bar.finish()

        logger.info("Getting the PCDs with the lowest voxel densities...")
        argsort = np.argsort(numbers_of_voxels)
        for index in argsort[0:20]:
            logger.info("%.2f : %.2f", numbers_of_voxels[index], self.pcd_paths[index])

        logger.info("Rendering histograms...")
        plt.hist(voxelgrid_sizes)
        plt.title("Distribution of voxelgrid-sizes.")
        plt.show()
        plt.close()

        plt.hist(numbers_of_voxels)
        plt.title("Distribution of number of voxels.")
        plt.show()
        plt.close()

        plt.hist(voxel_densities)
        plt.title("Distribution of voxel densities.")
        plt.show()
        plt.close()


def test_generator():

    if os.path.exists("datasetpath.txt"):
        dataset_path = open("datasetpath.txt", "r").read().replace("\n", "")
    else:
        dataset_path = "../data"

    data_generator = DataGenerator(dataset_path=dataset_path,
                                   input_type="pointcloud", output_targets=["height", "weight"])

    logger.info("jpg_paths %d", len(data_generator.jpg_paths))
    logger.info("pcd_paths %d", len(data_generator.jpg_paths))
    logger.info("json_paths_personal %d", len(data_generator.jpg_paths))
    logger.info("json_paths_measures %d", len(data_generator.jpg_paths))
    logger.info("QR-Codes: \n" + "\n".join(data_generator.qrcodes))
    #logger.info(data_generator.qrcodes_dictionary)
    data_generator.print_statistics()

    qrcodes_shuffle = list(data_generator.qrcodes)
    random.shuffle(qrcodes_shuffle)
    split_index = int(0.8 * len(qrcodes_shuffle))
    qrcodes_train = qrcodes_shuffle[:split_index]
    qrcodes_validate = qrcodes_shuffle[split_index:]

    logger.info("Training data:")
    x_train, y_train = next(data_generator.generate(
        size=200, qrcodes_to_use=qrcodes_train, verbose=True))
    logger.info(x_train.shape)
    logger.info(y_train.shape)
    logger.info("")

    logger.info("Validation data:")
    x_validate, y_validate = next(data_generator.generate(
        size=20, qrcodes_to_use=qrcodes_validate, verbose=True))
    logger.info(x_validate.shape)
    logger.info(y_validate.shape)
    logger.info("")


def test_dataset():

    if os.path.exists("datasetpath.txt"):
        dataset_path = open("datasetpath.txt", "r").read().replace("\n", "")
    else:
        dataset_path = "../data"

    data_generator = DataGenerator(dataset_path=dataset_path,
                                   input_type="image", output_targets=["height", "weight"])

    x_qrcodes, x_inputs, y_outputs = data_generator.generate_dataset(data_generator.qrcodes[0:4])
    logger.info(len(x_qrcodes))


def test_parameters():

    if os.path.exists("datasetpath.txt"):
        dataset_path = open("datasetpath.txt", "r").read().replace("\n", "")
    else:
        dataset_path = "../data"

    logger.info("Testing image...")
    data_generator = DataGenerator(dataset_path=dataset_path, input_type="image", output_targets=[
                                   "height", "weight"], image_target_shape=(20, 20))
    x_input, y_output = next(data_generator.generate(size=1))
    assert x_input.shape[1:3] == (20, 20)

    logger.info("Testing voxelgrid...")
    data_generator = DataGenerator(dataset_path=dataset_path, input_type="voxelgrid", output_targets=[
                                   "height", "weight"], voxelgrid_target_shape=(20, 20, 20))
    x_input, y_output = next(data_generator.generate(size=1))
    assert x_input.shape[1:] == (20, 20, 20)

    logger.info("Testing pointcloud...")
    data_generator = DataGenerator(dataset_path=dataset_path, input_type="pointcloud", output_targets=[
                                   "height", "weight"], pointcloud_target_size=16000)
    x_input, y_output = next(data_generator.generate(size=1))
    assert x_input.shape[1:] == (20, 4)

    logger.info("Done.")


def generate_data(class_self, size, qrcodes_to_use, verbose, yield_file_paths, output_queue):
    if verbose is True:
        logger.info("Generating QR-codes to be used: %s", qrcodes_to_use)

    assert size != 0

    x_inputs = []
    y_outputs = []
    file_paths = []

    if verbose is True:
        bar = progressbar.ProgressBar(max_value=size)
    while len(x_inputs) < size:

        # Get a random QR-code.
        qrcode = random.choice(qrcodes_to_use)

        # Get targets and paths randomly.
        if qrcode not in class_self.qrcodes_dictionary.keys():
            continue
        targets, jpg_paths, pcd_paths = random.choice(class_self.qrcodes_dictionary[qrcode])

        # No pointclouds.
        if len(pcd_paths) == 0:
            continue

        # Get a sample.
        x_input = None
        y_output = None
        file_path = None

        # Get the input. Not dealing with sequences.
        if class_self.sequence_length == 0:
            x_input, file_path = get_input(class_self, jpg_paths, pcd_paths)

        # Get the input. Dealing with sequences here.
        else:
            count = 0
            x_input, file_path = [], []
            while count != class_self.sequence_length:
                x, f = get_input(class_self, jpg_paths, pcd_paths)
                if x is not None and f is not None:
                    x_input.append(x)
                    file_path.append(f)
                    count += 1
            x_input = np.array(x_input)
            file_path = np.array(file_path)

        # Set the output.
        y_output = targets

        # Got a proper sample.
        if x_input is not None and y_output is not None and file_path is not None:
            x_inputs.append(x_input)
            y_outputs.append(y_output)
            file_paths.append(file_path)

        assert len(x_inputs) == len(y_outputs)
        assert len(y_outputs) == len(file_paths)

        if verbose is True:
            bar.update(len(x_inputs))

    if verbose is True:
        bar.finish()

    assert len(x_inputs) == size
    assert len(y_outputs) == size

    # Turn everything into ndarrays.
    x_inputs = np.array(x_inputs)
    y_outputs = np.array(y_outputs)

    # Prepare result values.
    assert len(x_inputs) == size
    assert len(y_outputs) == size
    if yield_file_paths is False:
        return_values = (x_inputs, y_outputs)
    else:
        return_values = (x_inputs, y_outputs, file_paths)

    # This is used in multiprocessing. Creates a pickle file and puts the data there.
    if output_queue is not None:
        output_path = uuid.uuid4().hex + ".p"
        pickle.dump(return_values, open(output_path, "wb"))
        output_queue.put(output_path)
    else:
        return return_values


def get_input(class_self, jpg_paths, pcd_paths):
    # Get a random image.
    if class_self.input_type == "image":
        if len(jpg_paths) == 0:
            logger.info("777")
            return None, None
        jpg_path = random.choice(jpg_paths)
        image = class_self._load_image(jpg_path)
        file_path = jpg_path
        x_input = image

    # Get a random voxelgrid.
    elif class_self.input_type == "voxelgrid":
        if len(pcd_paths) == 0:
            return None, None
        pcd_path = random.choice(pcd_paths)
        try:
            voxelgrid = class_self._load_voxelgrid(pcd_path)
            file_path = pcd_path
            x_input = voxelgrid
        except Exception as e:
            print(e)
            return None, None

    # Get a random pointcloud.
    elif class_self.input_type == "pointcloud":
        if len(pcd_paths) == 0:
            return None, None
        pcd_path = random.choice(pcd_paths)
        try:
            pointcloud = class_self._load_pointcloud(pcd_path)
            file_path = pcd_path
            x_input = pointcloud
        except Exception as e:
            print(e)
            return None, None

    # Should not happen.
    else:
        raise Exception("Unknown input_type: " + class_self.input_type)

    return x_input, file_path


def create_datagenerator_from_parameters(dataset_path, dataset_parameters):
    logger.info("Creating data-generator...")
    datagenerator = DataGenerator(
        dataset_path=dataset_path,
        input_type=dataset_parameters["input_type"],
        output_targets=dataset_parameters["output_targets"],
        sequence_length=dataset_parameters.get("sequence_length", 0),
        voxelgrid_target_shape=dataset_parameters.get("voxelgrid_target_shape", None),
        voxel_size_meters=dataset_parameters.get("voxel_size_meters", None),
        voxelgrid_random_rotation=dataset_parameters.get("voxelgrid_random_rotation", None),
        pointcloud_target_size=dataset_parameters.get("pointcloud_target_size", None),
        pointcloud_random_rotation=dataset_parameters.get("pointcloud_random_rotation", None)
    )
    #datagenerator.print_statistics()
    return datagenerator


def get_dataset_path():
    if os.path.exists("datasetpath.txt"):
        with open("datasetpath.txt", "r") as file:
            dataset_path = file.read().replace("\n", "")
    else:
        dataset_path = "../data"
    return dataset_path


if __name__ == "__main__":
    test_generator()
    #test_dataset()
    #test_parameters()



"""The code in this file was copied from: https://www.avanwyk.com/finding-a-learning-rate-in-tensorflow-2/"""

import matplotlib.pyplot as plt
import tensorflow as tf

from tensorflow.keras.callbacks import Callback


class LRFinder(Callback):
    """`Callback` that exponentially adjusts the learning rate after each training batch between `start_lr` and
    `end_lr` for a maximum number of batches: `max_step`. The loss and learning rate are recorded at each step allowing
    visually finding a good learning rate as per https://sgugger.github.io/how-do-you-find-a-good-learning-rate.html via
    the `plot` method.
    """

    def __init__(self, start_lr: float = 1e-7, end_lr: float = 10, max_steps: int = 100, smoothing=0.9):
        super(LRFinder, self).__init__()
        self.start_lr, self.end_lr = start_lr, end_lr
        self.max_steps = max_steps
        self.smoothing = smoothing
        self.step, self.best_loss, self.avg_loss, self.lr = 0, 0, 0, 0
        self.lrs, self.losses = [], []

    def on_train_begin(self, logs=None):
        self.step, self.best_loss, self.avg_loss, self.lr = 0, 0, 0, 0
        self.lrs, self.losses = [], []

    def on_train_batch_begin(self, batch, logs=None):
        self.lr = self.exp_annealing(self.step)
        tf.keras.backend.set_value(self.model.optimizer.lr, self.lr)

    def on_train_batch_end(self, batch, logs=None):
        logs = logs or {}
        loss = logs.get('loss')
        step = self.step
        if loss:
            self.avg_loss = self.smoothing * self.avg_loss + (1 - self.smoothing) * loss
            smooth_loss = self.avg_loss / (1 - self.smoothing ** (self.step + 1))
            self.losses.append(smooth_loss)
            self.lrs.append(self.lr)

            if step == 0 or loss < self.best_loss:
                self.best_loss = loss

            if smooth_loss > 4 * self.best_loss or tf.math.is_nan(smooth_loss):
                self.model.stop_training = True

        if step == self.max_steps:
            self.model.stop_training = True

        self.step += 1

    def exp_annealing(self, step):
        return self.start_lr * (self.end_lr / self.start_lr) ** (step * 1. / self.max_steps)

    def plot(self):
        fig, ax = plt.subplots(1, 1)
        ax.set_ylabel('Loss')
        ax.set_xlabel('Learning Rate (log scale)')
        ax.set_xscale('log')
        ax.xaxis.set_major_formatter(plt.FormatStrFormatter('%.0e'))
        ax.plot(self.lrs, self.losses)



"""To identify the child from the RGB images.

Use the pytorch Mask R-CNN Resnet50 library to identify the child
and then using the mask, applied binary image-segmentation to
represent the child pixel as '1' and background pixel as '0'
Further, calculating the mask area and the percentage of
body pixels to total image pixels
"""
import time
import logging

import numpy as np
from torchvision.models.detection import maskrcnn_resnet50_fpn

from cgmml.common.reliability.background_segm.imgseg.predict import predict

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)

model = maskrcnn_resnet50_fpn(pretrained=True)


def predict_by_resize(image, factor=10):
    """Applied MaskRCNN on downscaled image, by default the factor is 10x."""
    logger.info("Resizing image by %d x", factor)
    newsize = (int(image.size[0] / factor), int(image.size[1] / factor))
    logger.info("Resized Dimension %s", newsize)
    start_time = time.time()
    out = predict(image.resize(newsize), model)
    logger.info("Time: %s s", time.time() - start_time)

    # Binary Image Segmentation
    threshold = 0.5
    masks = out['masks'][0][0]
    masks = masks > threshold
    out['masks'][0][0] = masks.astype(int)

    return out


def get_mask_information(segmented_image):
    """Return the mask information."""
    width = len(segmented_image['masks'][0][0][0])
    height = len(segmented_image['masks'][0][0])

    # Get the masked area
    mask_area = int(np.reshape(
        segmented_image['masks'],
        (-1, segmented_image['masks'].shape[-1])).astype(np.float32).sum())

    # Get mask stats like percentage of body coverage of total area & mask area
    perc_body_covered = (mask_area * 100) / (width * height)
    perc_body_covered = round(perc_body_covered, 2)
    logger.info("Mask Area: %.2f px", mask_area)
    logger.info("Percentage of body pixels to total img pixels: %.2f %", perc_body_covered)
    return mask_area, perc_body_covered



from pathlib import Path

REPO_DIR = Path(__file__).parents[1].absolute()
WHO_TABLES = [
    'wfl_boys_0_2_zscores.json', 'wfl_girls_0_2_zscores.json',
    'wfh_boys_2_5_zscores.json', 'wfh_girls_2_5_zscores.json',
    'wfa_boys_0_5_zscores.json', 'wfa_girls_0_5_zscores.json',
    'lhfa_boys_0_5_zscores.json', 'lhfa_girls_0_5_zscores.json',
]



import logging
import random
import shutil
from importlib import import_module
from pathlib import Path

import tensorflow as tf
from azureml.core.run import Run

from constants import REPO_DIR

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)


def copy_dir(src: Path, tgt: Path, glob_pattern: str, should_touch_init: bool = False):
    logger.info("Creating temp folder")
    if tgt.exists():
        shutil.rmtree(tgt)
    tgt.mkdir(parents=True, exist_ok=True)
    if should_touch_init:
        (tgt / '__init__.py').touch(exist_ok=False)

    paths_to_copy = list(src.glob(glob_pattern))
    logger.info(f"Copying to {tgt} the following files: {str(paths_to_copy)}")
    for p in paths_to_copy:
        destpath = tgt / p.relative_to(src)
        destpath.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(p, destpath)


def is_offline_run(run: Run) -> bool:
    return run.id.startswith("OfflineRun")


# Get the current run.
RUN = Run.get_context()

from cgmml.common.evaluation.eval_utilities import (  # noqa: E402, F401
    is_ensemble_evaluation, is_multiartifact_evaluation)
from cgmml.common.evaluation.evaluation_classes import (  # noqa: E402, F401
    Evaluation, EnsembleEvaluation, MultiartifactEvaluation)
from cgmml.common.model_utils.run_initialization import OfflineRunInitializer, OnlineRunInitializer  # noqa: E402

logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d')


QA_CONFIG_MODULES = [
    'qa_config_height',  # takes 14min in CI
    'qa_config_height_clean',  # takes 9min in CI
    # 'qa_config_height_dropout',  # takes 14min in CI
    # 'qa_config_height_deep_ensemble',  # takes 50min in CI
    # 'qa_config_height_mcnn',  # takes 19min in CI
    # 'qa_config_height_filter',
    'qa_config_weight',  # takes 13min in CI
    # 'qa_config_weight_dropout',  # takes 13min in CI
]

if __name__ == "__main__":

    for qa_config_module in QA_CONFIG_MODULES:
        qa_config = import_module(qa_config_module)
        logger.info('Using the following config: %s', qa_config_module)

        model_config = qa_config.MODEL_CONFIG
        eval_config = qa_config.EVAL_CONFIG
        data_config = qa_config.DATA_CONFIG
        result_config = qa_config.RESULT_CONFIG
        filter_config = qa_config.FILTER_CONFIG if getattr(qa_config, 'FILTER_CONFIG', False) else None

        # Make experiment reproducible
        tf.random.set_seed(eval_config.SPLIT_SEED)
        random.seed(eval_config.SPLIT_SEED)

        if is_offline_run(RUN):
            output_csv_path = str(REPO_DIR / 'data' / result_config.SAVE_PATH)
            initializer = OfflineRunInitializer(data_config, eval_config)
        else:
            output_csv_path = result_config.SAVE_PATH
            initializer = OnlineRunInitializer(data_config, eval_config, RUN)

        if is_ensemble_evaluation(model_config):
            model_base_dir = (REPO_DIR / 'data' / model_config.EXPERIMENT_NAME) if is_offline_run(RUN) else Path('.')
            eval_class = EnsembleEvaluation
            descriptor = model_config.EXPERIMENT_NAME
        else:
            model_base_dir = REPO_DIR / 'data' / model_config.RUN_ID if is_offline_run(RUN) else Path('.')
            eval_class = MultiartifactEvaluation if is_multiartifact_evaluation(data_config) else Evaluation
            descriptor = model_config.RUN_ID
        evaluation = eval_class(model_config, data_config, model_base_dir, initializer.dataset_path)
        evaluation.get_the_model_path(initializer.workspace)

        # Get the QR-code paths
        qrcode_paths = evaluation.get_the_qr_code_path()
        if getattr(eval_config, 'DEBUG_RUN', False) and len(qrcode_paths) > eval_config.DEBUG_NUMBER_OF_SCAN:
            qrcode_paths = qrcode_paths[:eval_config.DEBUG_NUMBER_OF_SCAN]
            logger.info("Executing on %d qrcodes for FAST RUN", eval_config.DEBUG_NUMBER_OF_SCAN)

        dataset_evaluation, paths_belonging_to_predictions = evaluation.prepare_dataset(qrcode_paths, filter_config)
        prediction_array = evaluation.get_prediction_(evaluation.model_path_or_paths, dataset_evaluation)
        logger.info("Prediction made by model on the depthmaps...")
        logger.info(prediction_array)

        df = evaluation.prepare_dataframe(paths_belonging_to_predictions, prediction_array, result_config)
        evaluation.evaluate(df, result_config, eval_config, output_csv_path, descriptor)

        # Done.
        initializer.run.complete()



#
# Child Growth Monitor - Free Software for Zero Hunger
# Copyright (c) 2019 Dr. Christian Pfitzner <christian.pfitzner@th-nuernberg.de> for Welthungerhilfe
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os
from enum import IntEnum

import numpy as np
import pandas as pd
from cv2 import cv2
from pyntcloud import PyntCloud
from pyntcloud.io import write_ply

from cgmml.common.rgbd_toolkit.cgm_fusion.calibration import (get_extrinsic_matrix_depth,
                                                              get_intrinsic_matrix_depth, get_k_depth)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)

HEIGHT = 224
WIDTH = 172


def fuse_point_cloud(points, rgb_vals, confidence, seg_vals):
    df = pd.DataFrame(columns=['x', 'y', 'z', 'red', 'green', 'blue',
                               'seg'])  # 'nx', 'ny', 'nz'])

    df['x'] = points[:, 0]  # saving carthesian coordinates
    df['y'] = points[:, 1]
    df['z'] = points[:, 2]

    df['red'] = rgb_vals[:, 2].astype(np.uint8)  # saving the color
    df['green'] = rgb_vals[:, 1].astype(np.uint8)
    df['blue'] = rgb_vals[:, 0].astype(np.uint8)

    df['c'] = confidence[:].astype(np.float)  # saving the confidence

    df['seg'] = seg_vals[:].astype(np.float)  # saving the segmentation

    # df['nx']    = normals[:, 0]                             # normal x coordinate
    # df['ny']    = normals[:, 1]                             # normal y coordinate
    # df['nz']    = normals[:, 2]                             # normal z coordinate

    new_pc = PyntCloud(df)
    return new_pc


def write_color_ply(fname, points, color_vals, confidence, normals):
    new_pc = fuse_point_cloud(points, color_vals, confidence, normals)
    write_ply(fname, new_pc.points, as_text=True)
    logger.info(fname)


def apply_projection(points, calibration_file):
    intrinsic = get_intrinsic_matrix_depth(calibration_file)

    ext_d = get_extrinsic_matrix_depth(calibration_file, idx=4)

    r_vec = ext_d[:3, :3]
    t_vec = -ext_d[:3, 3]

    k1, k2, k3 = get_k_depth(calibration_file)
    im_coords, _ = cv2.projectPoints(points, r_vec, t_vec, intrinsic[:3, :3],
                                     np.array([k1, k2, 0, 0]))

    return im_coords


class Channel(IntEnum):
    x = 0
    y = 1
    z = 2
    confidence = 3
    red = 4
    green = 5
    blue = 6
    segmentation = 7
    nx = 8
    ny = 9
    nz = 10


def get_depth_channel(ply_path, output_path_np, output_path_png, calibration_file):
    if not os.path.exists(calibration_file):  # check if the califile exists
        logger.error('Calibration does not exist')
        return

    # get a default black image
    nr_of_channels = 1
    viz_image = np.zeros((HEIGHT, WIDTH, nr_of_channels), np.float64)

    try:
        cloud = PyntCloud.from_file(ply_path)  # load the data from the files
    except ValueError as e:
        logger.error("Error reading point cloud")
        logger.error(str(e))
        logger.error(ply_path)

    logger.info(cloud.points.values[:, 3])
    r = cloud.points.values[:, 3]
    logger.info(min(r))
    logger.info(max(r))
    logger.info(np.mean(r))

    points = cloud.points.values[:, :3]  # get x y z
    z = cloud.points.values[:, 2]  # get only z coordinate
    z = (z - min(z)) / (max(z) - min(z))  # normalize the data to 0 to 1

    # iterat of the points and calculat the x y coordinates in the image
    # get the data for calibration
    im_coords = apply_projection(points, calibration_file)

    # manipulate the pixels color value depending on the z coordinate
    # TODO make this a function
    for i, t in enumerate(im_coords):
        x, y = t.squeeze()
        x = int(np.round(x))
        y = int(np.round(y))
        if x >= 0 and x < HEIGHT and y >= 0 and y < WIDTH:
            viz_image[x, y] = z[i]  # 255 #255-255*z[i]

    # img_debug = cv2.normalize(src=viz_image,
    #                           dst=None,
    #                           alpha=0,
    #                           beta=255,
    #                           norm_type=cv2.NORM_MINMAX,
    #                           dtype=cv2.CV_8U)

    cv2.imwrite("/tmp/viz_debug.png", viz_image)

    # resize and  return the image after pricessing
    # dim = (180, 240)
    # viz_image = cv2.resize(viz_image, dim, interpolation = cv2.INTER_AREA)

    #np.save(output_path_np, viz_image)
    #viz_2 = np.load("/tmp/out.npy")

    img_n = cv2.normalize(src=viz_image,
                          dst=None,
                          alpha=0,
                          beta=255,
                          norm_type=cv2.NORM_MINMAX,
                          dtype=cv2.CV_8U)
    cv2.imwrite(output_path_png, img_n)

    return viz_image


def get_rgbd_channel(ply_path, output_path_np, calibration_file):
    if not os.path.exists(calibration_file):  # check if the califile exists
        logger.error('Calibration does not exist')
        return

    # get a default black image
    nr_of_channels = 4
    viz_image = np.zeros((HEIGHT, WIDTH, nr_of_channels), np.float64)

    try:
        cloud = PyntCloud.from_file(ply_path)  # load the data from the files
    except ValueError as e:
        logger.error(" Error reading point cloud ")
        logger.error(str(e))
        logger.error(ply_path)

    points = cloud.points.values[:, :3]  # get x y z
    z = cloud.points.values[:, 2]  # get only z coordinate
    z = (z - min(z)) / (max(z) - min(z))  # normalize the data to 0 to 1
    r = cloud.points.values[:, 4]
    g = cloud.points.values[:, 5]
    b = cloud.points.values[:, 6]

    # iterat of the points and calculat the x y coordinates in the image
    # get the data for calibration
    im_coords = apply_projection(points, calibration_file)
    #im_coords=np.nan_to_num(im_coords)                     #removing nans

    # manipulate the pixels color value depending on the z coordinate
    # TODO make this a function

    for i, t in enumerate(im_coords):
        x, y = t.squeeze()
        x = int(np.round(x))
        y = int(np.round(y))
        if x >= 0 and x < HEIGHT and y >= 0 and y < WIDTH:
            viz_image[x, y, 0] = r[i]
            viz_image[x, y, 1] = g[i]
            viz_image[x, y, 2] = b[i]
            viz_image[x, y, 3] = z[i]

    np.save(output_path_np, viz_image)
    return viz_image


def get_all_channel(ply_path, output_path_np, calibration_file):
    if not os.path.exists(calibration_file):  # check if the califile exists
        logger.error('Calibration does not exist')
        return

    # get a default black image
    nr_of_channels = 11
    viz_image = np.zeros((HEIGHT, WIDTH, nr_of_channels), np.float64)

    try:
        cloud = PyntCloud.from_file(ply_path)  # load the data from the files
    except ValueError as e:
        logger.error(" Error reading point cloud ")
        logger.error(str(e))
        logger.error(ply_path)

    points = cloud.points.values[:, :3]  # get x y z
    x = cloud.points.values[:, 0]
    y = cloud.points.values[:, 1]
    z = cloud.points.values[:, 2]  # get only z coordinate
    z = (z - min(z)) / (max(z) - min(z))  # normalize the data to 0 to 1

    r = cloud.points.values[:, 4]
    g = cloud.points.values[:, 5]
    b = cloud.points.values[:, 6]

    seg = cloud.points.values[:, 7]
    conf = cloud.points.values[:, 3]

    # nx = cloud.points.values[:, 8]
    # nx = cloud.points.values[:, 9]
    # nz = cloud.points.values[:, 10]

    # iterat of the points and calculat the x y coordinates in the image
    # get the data for calibration
    im_coords = apply_projection(points)

    # manipulate the pixels color value depending on the z coordinate
    # TODO make this a function
    for i, t in enumerate(im_coords):
        x, y = t.squeeze()
        x = int(np.round(x))
        y = int(np.round(y))
        if x >= 0 and x < HEIGHT and y >= 0 and y < WIDTH:
            viz_image[x, y, 0] = x[i]
            viz_image[x, y, 1] = y[i]
            viz_image[x, y, 2] = z[i]

            viz_image[x, y, 3] = r[i]
            viz_image[x, y, 4] = g[i]
            viz_image[x, y, 5] = b[i]

            viz_image[x, y, 6] = seg[i]
            viz_image[x, y, 7] = conf[i]

            # viz_image[x, y, 8] = nx[i]
            # viz_image[x, y, 9] = ny[i]
            # viz_image[x, y, 10] = nz[i]

    np.save(output_path_np, viz_image)
    return viz_image


def get_viz_channel(calibration_file,
                    ply_path,
                    channel=Channel.z,
                    output_path="/tmp/output.png"):

    if not os.path.exists(calibration_file):  # check if the califile exists
        logger.error('Calibration does not exist')
        return

    # get a default black image
    nr_of_channels = 1
    viz_image = np.zeros((HEIGHT, WIDTH, nr_of_channels), np.uint8)

    # get the points from the pointcloud
    try:
        cloud = PyntCloud.from_file(ply_path)  # load the data from the files
    except ValueError as e:
        logger.error(" Error reading point cloud ")
        logger.error(str(e))

    # logger.info(int(channel))

    points = cloud.points.values[:, :3]  # get x y z
    z = cloud.points.values[:, int(channel)]  # get only z coordinate
    z = (z - min(z)) / (max(z) - min(z))  # normalize the data to 0 to 1

    # iterat of the points and calculat the x y coordinates in the image
    # get the data for calibration
    im_coords = apply_projection(points, calibration_file)

    # manipulate the pixels color value depending on the z coordinate
    # TODO make this a function
    for i, t in enumerate(im_coords):
        x, y = t.squeeze()
        x = int(np.round(x))
        y = int(np.round(y))
        if x >= 0 and x < HEIGHT and y >= 0 and y < WIDTH:
            viz_image[x, y] = 255 * z[i]

    # resize and  return the image after pricessing
    #imgScale = 0.25
    #newX, newY = viz_image.shape[1] * imgScale, viz_image.shape[0] * imgScale
    cv2.imwrite(output_path, viz_image)

    return viz_image


def get_viz_rgb(ply_path):
    """Function to get the rgb from a point cloud as an image for visualization"""
    get_viz_channel(ply_path, channel=Channel.red, output_path="/tmp/red.png")


def get_viz_confidence(ply_path):
    """Function to get the confidence from a point cloud as an image for visualization"""
    get_viz_channel(ply_path,
                    channel=Channel.confidence,
                    output_path="/tmp/confidence.png")


def get_viz_depth(ply_path):
    get_viz_channel(ply_path, channel=Channel.z, output_path="/tmp/depth.png")


def get_viz_segmentation(ply_path):
    """Function to get the segmentation from a point cloud as an image for visualization"""
    get_viz_channel(ply_path,
                    channel=Channel.segmentation,
                    output_path="/tmp/segmentation.png")


def get_viz_normal_z(ply_path):
    get_viz_channel(ply_path,
                    channel=Channel.nz,
                    output_path="/tmp/normals.png")



def read_csv(filepath: str) -> list:
    output = []
    with open(filepath) as f:
        for line in f:
            data = line.replace("\n", "")
            output.append(data.split(','))
    return output


def write_csv(filepath: str, data: list):
    with open(filepath, "w") as f:
        for row in data:
            line = ''
            for value in row:
                if len(line) > 0:
                    line = line + ','
                line = line + str(value)
            f.write(line + '\n')



import zipfile
import pickle
import multiprocessing
import json
import logging
from typing import Tuple

import numpy as np
import cv2 as cv
from glob2 import glob
from pathlib import Path
from glob import glob1

TARGET_HEIGHT = 180
TARGET_WIDTH = 240
TARGET_PATH = '/mnt/huawei_dataset/anon-rgbd-5kscans'
SOURCE_PATH = '/mnt/huawei_dataset/huawei_data'

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)


def load_depth(fpath: str) -> Tuple[bytes, int, int, float, float]:
    """Take ZIP file and extract depth and metadata
    Args:
        fpath (str): File path to the ZIP
    Returns:
        depth_data (bytes): depthmap data
        width(int): depthmap width in pixel
        height(int): depthmap height in pixel
        depth_scale(float)
        max_confidence(float)
    """
    with zipfile.ZipFile(fpath) as z:
        with z.open('data') as f:
            # Example for a first_line:
            # '180x135_0.001_7_0.57045287_-0.0057296_0.0022602521_0.82130724_-0.059177425_0.0024800065_0.030834956'
            first_line = f.readline().decode().strip()

            file_header = first_line.split("_")

            # header[0] example: 180x135
            width, height = file_header[0].split("x")
            width, height = int(width), int(height)
            depth_scale = float(file_header[1])
            max_confidence = float(file_header[2])

            depth_data = f.read()
    return depth_data, width, height, depth_scale, max_confidence


def prepare_depthmap(data: bytes, width: int, height: int, depth_scale: float) -> np.ndarray:
    """Convert bytes array into np.array"""
    output = np.zeros((width, height, 1))
    for cx in range(width):
        for cy in range(height):
            # depth data scaled to be visible
            output[cx][height - cy - 1] = parse_depth(cx, cy, data, depth_scale, width)
    arr = np.array(output, dtype='float32')
    return arr.reshape(width, height)


def parse_depth(tx: int, ty: int, data: bytes, depth_scale: float, width: int) -> float:
    assert isinstance(tx, int)
    assert isinstance(ty, int)

    depth = data[(ty * width + tx) * 3 + 0] << 8
    depth += data[(ty * width + tx) * 3 + 1]

    depth *= depth_scale
    return depth


def check_correspondence(depth_frame, rgb_file_list):
    depth_name = depth_frame.split('.depth')[0]
    frame_name = depth_name.split('depth_')
    rgb_frame = f'rgb_{frame_name[-1]}.jpg'
    assert rgb_frame in rgb_file_list
    return depth_frame, rgb_frame


def image_resize(image_path):
    image = cv.imread(image_path)
    image_rotated = cv.rotate(image, cv.ROTATE_90_CLOCKWISE)
    image_rotated = cv.resize(image_rotated, (TARGET_HEIGHT, TARGET_WIDTH))
    return image_rotated


def read_json(filepath):
    with open(filepath) as json_data:
        label_data = json.load(json_data)
    return label_data


def process_depthmap(depthmaps):
    split_dirpath = depthmaps.split('/depth')[0]
    json_fpath = f'{split_dirpath}/targets.json'
    label_data = read_json(json_fpath)
    labels = np.array([label_data['height'], label_data['weight'],
                       label_data['muac'], label_data['age'], label_data['sex']])
    qrcode = depthmaps.split('/')[5]
    qrcode_dirpath = f'{TARGET_PATH}/{qrcode}'
    Path(qrcode_dirpath).mkdir(parents=True, exist_ok=True)
    rgb_dirpath = f'{split_dirpath}/rgb'
    rgb_list = glob1(rgb_dirpath, '*.jpg')
    abs_depth_pattern = f'{depthmaps}/*.depth'
    depthmap_files = glob(abs_depth_pattern)
    for unique_depthmaps in depthmap_files:
        try:
            depthmap_image_path, image_path = check_correspondence(unique_depthmaps, rgb_list)
        except Exception as e:
            message = f" Error '{e.message}' occurred. Depthmap file '{depthmap_files}' not found."
            logger.info(message)
            continue
        scan_type = image_path.split('_')[3]
        artifact_name = image_path.split('rgb_')[1]
        scan_type_dirpath = f'{qrcode_dirpath}/{scan_type}'
        pickle_file = artifact_name.replace('.jpg', '.p')
        full_fpath = f'{scan_type_dirpath}/{pickle_file}'
        Path(scan_type_dirpath).mkdir(parents=True, exist_ok=True)
        data, width, height, depth_scale, _ = load_depth(depthmap_image_path)
        depthmap_huawei = prepare_depthmap(data, width, height, depth_scale)
        image_full_fpath = f'{rgb_dirpath}/{image_path}'
        resized_image = image_resize(image_full_fpath)
        pickled_data = (resized_image, depthmap_huawei, labels)
        pickle.dump(pickled_data, open(full_fpath, "wb"))


if __name__ == "__main__":
    source_pattern = f'{SOURCE_PATH}'
    dataset_list = []
    proc = multiprocessing.Pool()
    for elem in Path(source_pattern).rglob('*/depth'):
        proc.apply_async(dataset_list.append(elem))
    proc.close()
    proc.join()  # Wait for all child processes to close
    proc = multiprocessing.Pool()
    for depthimages in dataset_list:
        # process_depthmap(depthimages)
        # launch a process for each file (ish).
        # The result will be approximately one process per CPU core available.
        proc.apply_async(process_depthmap, [depthimages])

    proc.close()
    proc.join()  # Wait for all child processes to close



import argparse
import concurrent.futures
import datetime
import logging
import os
import pickle
import sys
import warnings
from pathlib import Path

import azureml.core
import numpy as np
import pandas as pd
from PIL import Image
from tqdm import tqdm

from cgmml.common.rgbd_toolkit.cgm_fusion.fusion import fuse_rgbd
from cgmml.common.rgbd_toolkit.get_timestamps import get_timestamps_from_pcd, get_timestamps_from_rgb

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)

sys.path.append('../cgm-ml')
sys.path.append(os.path.dirname(os.getcwd()))

# check core SDK version number
logger.info("Azure ML SDK Version: ", azureml.core.VERSION)

warnings.filterwarnings("ignore")


def find_closest(rgb, pcd):
    """Find the closest rgbd images for a given pcd

    rgb must be sorted
    """
    idx = rgb.searchsorted(pcd)
    idx = np.clip(idx, 1, len(rgb) - 1)
    left = rgb[idx - 1]
    right = rgb[idx]
    idx -= pcd - left < right - pcd
    return idx


def get_filename(pcd_file, rgbd_folder, qr_folder):

    # save the new data to the folder
    fused_folder, pc_filename = os.path.split(str(pcd_file))
    pcd_path_old = pcd_file

    # finding if its standing or lying artifact
    scan_type = pcd_file.split("/")[-1].split("_")[3]

    # replace the pcd and the pc_ in the path for fused data
    pc_filename = pcd_path_old.replace(".pcd", ".ply")
    pc_filename = pc_filename.replace("pc_", "pcrgb_")

    # make a folder for scan type
    qr_rgbd = os.path.join(rgbd_folder, qr_folder)
    qr_rgbd_scan = os.path.join(qr_rgbd, str(scan_type))

    # make the output rgbd filename
    rgbd_filename = pc_filename.replace(
        fused_folder.split("measure")[0], qr_rgbd_scan + "/")

    # manipulate the filename
    rgbd_filename = rgbd_filename.replace(".ply", ".pkl")
    rgbd_filename = rgbd_filename.replace("/pc/", "")

    # check if output rgbd folder exists
    rgbd_folder_ = os.path.dirname(rgbd_filename)
    if not os.path.isfile(rgbd_folder_):
        logger.info("Folder does not exist for " + str(rgbd_filename))
        os.makedirs(rgbd_folder_, exist_ok=True)
        logger.info("Created folder " + str(rgbd_folder_))
    return rgbd_filename


def process_pcd(paths, process_index=0):

    pcd_file = paths[0]
    jpg_file = paths[1]

    #rgb image
    image = Image.open(jpg_file)

    #get the qr folder
    qr_folder = str(Path(qr)).split("/")[-1]

    calibration_file = "./calibration.xml"

    if args.pickled:
        #get height and weight label for the corresponding artifact
        height = int(artifacts_file.loc[np.where(
            artifacts_file["qrcode"] == qr_folder)].iloc[0].loc["height"])
        weight = int(artifacts_file.loc[np.where(
            artifacts_file["qrcode"] == qr_folder)].iloc[0].loc["weight"])

    rgbd_filename = get_filename(pcd_file, rgbd_folder, qr_folder)

    logger.info("Going to writing new fused data to: " + rgbd_filename)

    #saving the rgbd file with labels as pickled data
    try:
        rgbdseg_arr = fuse_rgbd(calibration_file, pcd_file, image)  # , seg_path)
        if args.pickled:
            labels = np.array([height, weight])
            if not labels:
                logger.info("labels dont exist in artifacts.csv..exiting")
                sys.exit()
            data = (rgbdseg_arr, labels)
            pickle.dump(data, open(rgbd_filename, "wb"))
        else:
            #saving as a png file if not pickled data
            rgbd_filename = rgbd_filename.replace(".pkl", ".npy")
            np.save(rgbd_filename, rgbdseg_arr)

        logger.info("successfully wrote new data to" + rgbd_filename)
    except Exception as e:
        logger.error("Something went wrong.Skipping this file")
        logger.error(str(e))


def get_files(norm_rgb_time, rgb_path, norm_pcd_time, pcd_path):
    rgb_dict = {}

    for time, path in zip(norm_rgb_time, rgb_path):
        rgb_dict[time] = path
    sorted_rgb_dict = dict(sorted(rgb_dict.items()))
    norm_rgb_time = list(sorted_rgb_dict.keys())
    rgb_path = list(sorted_rgb_dict.values())
    norm_rgb_time = np.asarray(norm_rgb_time)
    files = []
    if len(norm_rgb_time) == 0:
        logger.info("no rgb images found")
        return []

    if len(norm_pcd_time) == 0:
        logger.info("no pcd images found")
        return []

    for i, pcd in enumerate(norm_pcd_time):
        nn = find_closest(norm_rgb_time, pcd)

        # get the original file path
        path, filename = os.path.split(str(pcd_path[i]))

        #located the nearest jpg file of that pcd
        pcd_file = pcd_path[i]
        jpg_file = rgb_path[nn]

        files.append([pcd_file, jpg_file])

    return files


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Give in the qrcode folder to get rgbd data')
    parser.add_argument('--input', required=True,
                        metavar='inputpath',
                        type=str,
                        help='the path to the input qrcode folder')
    parser.add_argument('--output', required=True,
                        metavar='outputpath',
                        default="./output_rgbd",
                        type=str,
                        help='output rgbd folder path')
    parser.add_argument("--num_workers",
                        metavar="workers",
                        type=int,
                        default=None,
                        help="no. of cpu workers you want to process with")
    parser.add_argument("--debug",
                        action='store_true',
                        help="Flag to enable debug mode")
    parser.add_argument(
        "--pickled",
        action='store_true',
        help="if you are processing on data of a dataset whose labels you have added into artifacts.csv")
    args = parser.parse_args()

    start = datetime.datetime.now()

    #reading artifacts.csv for pickled qrcode paths
    if args.pickled:
        artifacts = os.path.join(
            os.path.dirname(os.path.dirname(os.getcwd())),
            'data_utils/dataset_EDA/50k_pcd/artifacts.csv')
        mnt = args.input + "/qrcode/"
        artifacts_file = pd.read_csv(artifacts)
        qrcode = artifacts_file["qrcode"]
        paths = artifacts_file["storage_path"]
        unique_qr_codes = [os.path.join(mnt, line) for line in qrcode]

    else:
        folders = os.listdir(args.input)
        unique_qr_codes = [os.path.join(args.input, x) for x in folders if os.path.isdir(os.path.join(args.input, x))]

    ##validation of input directory
    if not os.path.exists(unique_qr_codes[0]):
        logger.info("Error:invalid input paths..exiting")
        sys.exit()

    #making output dir for storing rgbd files
    rgbd_folder = args.output
    if not os.path.exists(rgbd_folder):
        os.mkdir(rgbd_folder)

    #initialize empty lists for rgb and pcd paths of a given qr code
    rgb_paths = []
    pcd_paths = []

    #read jpg and pcd files of each qr code in the input directory.
    logger.info("Processing..")
    start = datetime.datetime.now()
    for qr in set(unique_qr_codes):
        logger.info("reading qr code" + str(qr))
        for dirname, dirs, qr_paths in os.walk(Path(qr)):
            for file in qr_paths:
                dir_path = os.path.join(dirname, file)
                if file.endswith(".jpg"):
                    rgb_paths.append(dir_path)
                if file.endswith(".pcd"):
                    pcd_paths.append(dir_path)

    #getting the timestamps of rgb and pcd paths
        [norm_rgb_time, rgb_path] = get_timestamps_from_rgb(rgb_paths)
        [norm_pcd_time, pcd_path] = get_timestamps_from_pcd(pcd_paths)

        paths = get_files(norm_rgb_time, rgb_path, norm_pcd_time, pcd_path)

        #processing every pcd file with its nearest rgb using multiprocessing workers
        if args.debug:
            logger.info("Debug")
            for path in paths:
                process_pcd(path)
        else:
            logger.info("Multiprocessing")
            with concurrent.futures.ProcessPoolExecutor(
                    max_workers=args.num_workers) as executor:
                res = list(tqdm(executor.map(process_pcd, paths),
                                total=len(paths)))

    end = datetime.datetime.now()
    diff = end - start
    logger.info("***Done***")
    logger.info("total time took is %.2f", diff)
